import numpy as np

import jax
import jax.numpy as jnp
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from jax.experimental import mesh_utils
import jaxopt

import supergrad.utils.fidelity as fidelity_lib


def get_sharding(*partition_spec):
    """Get multi-devices with sharding.

    Args:
        data: `jax.numpy.ndarray`
            The data to be distributed batched.
        partition_spec: `jax.sharding.PartitionSpec`
            The partition specification. The sharding is done along the "p"
            axis_name.
    """
    devices = mesh_utils.create_device_mesh((jax.local_device_count(),))
    mesh = Mesh(devices, 'p')
    return NamedSharding(mesh, PartitionSpec(*partition_spec))


def distributed_state_fidelity(target_states, computed_states):
    """Compute the state fidelity or the composite state (unitary) fidelity.
    The fidelity could be computed in a multi-devices cluster with the same sharding.

    Args:
        target_states: The target states.
        computed_states: The computed states.
    """
    # assert target_states.sharding == computed_states.sharding
    ar_state_fidelity = jnp.sum(jnp.conj(target_states) * computed_states,
                                axis=0)
    return jnp.abs(ar_state_fidelity.mean())**2


def distributed_overlap_with_auto_vz_compensation(psi_target,
                                                  psi_computed,
                                                  opt_method: str = 'gd',
                                                  options: dict = None):
    """
    Compute inner product of states, allows for vz compensation after.

    Args:
        psi_target (array): Target state
        psi_computed (array): State
        opt_method (string): optimizer, set `gd` for using the Gradient Descent
            solver; set `lbfgs` for using the LBFGS solver.
        options (dict): additional keyword arguments to be passed to the `jaxopt`
            optimizer.
    """
    n_qubit = round(np.log2(psi_target.shape[0]))
    if options is None:
        options = {}
    # initialize single qubit compensation
    initial_guess = jnp.zeros([n_qubit])

    def error(params, psi_target, psi_computed):
        psi2 = fidelity_lib.apply_nz(params, psi_computed)
        return jnp.log(1 - distributed_state_fidelity(psi_target, psi2))

    if opt_method == 'lbfgs':
        # call jaxopt differentiable LBFGS solver
        solver = jaxopt.LBFGS(error, **options)
    elif opt_method == 'gd':
        solver = jaxopt.GradientDescent(error, **options)
    else:
        raise ValueError('opt_method should be `gd` or `lbfgs`.')

    res = solver.run(initial_guess,
                     psi_target=psi_target,
                     psi_computed=psi_computed)
    # print(initial_guess)
    # print(res.state)
    # print(res.params)
    psi_optimal = fidelity_lib.apply_nz(res.params, psi_computed)

    return distributed_state_fidelity(psi_target, psi_optimal), psi_optimal


def distributed_fidelity_with_auto_vz_compensation(u_target,
                                                   u_computed,
                                                   opt_method: str = 'gd',
                                                   options: dict = None):
    """
    Compute fidelity, allows for vz compensation before/after.

    Args:
        u_target (array): Target unitary
        u_computed (array): Unitary
        opt_method (string): optimizer, set `gd` for using the Gradient Descent
            solver; set `lbfgs` for using the LBFGS solver.
        options (dict): additional keyword arguments to be passed to the `jaxopt`
            optimizer.
    """
    n_qubit = round(np.log2(u_target.shape[0]))
    if options is None:
        options = {}
    # initialize single qubit compensation
    initial_guess = jnp.zeros([2 * n_qubit])

    def error(params, u_target, u_computed):
        u2 = fidelity_lib.apply_2nz(params, u_computed)
        return jnp.log(1 - distributed_state_fidelity(u_target, u2))

    if opt_method == 'lbfgs':
        # call jaxopt differentiable LBFGS solver
        solver = jaxopt.LBFGS(error, **options)
    elif opt_method == 'gd':
        solver = jaxopt.GradientDescent(error, **options)
    else:
        raise ValueError('opt_method should be `gd` or `lbfgs`.')

    res = solver.run(initial_guess, u_target=u_target, u_computed=u_computed)
    # print(initial_guess)
    # print(res.state)
    # print(res.params)
    u_optimal = fidelity_lib.apply_2nz(res.params, u_computed)

    return distributed_state_fidelity(u_target, u_optimal), u_optimal



from collections import deque
import jax
import jax.numpy as jnp
import networkx as nx

from supergrad.scgraph.graph import SCGraph

# Below we try to specify the structure and parameters of a quantum processor
# This is done step by step.

# A detail explanation of all parameters in the resulting graph can be found in the end of this file.
# The explanation is roughly in the form of json schemas.

# For the examples in the notebook, we use parameters in PeriodicGraphOpt and its child class
# The parameters of PeriodicGraphOpt are obtained through some optimization
# This will be explained in a related paper.


class PeriodicGraph(SCGraph):
    """Device parameters for 5x5 periodic quantum processor graph.


    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(None)
        # initialize graph
        temp_graph = nx.grid_2d_graph(5, 5, periodic=True)

        params_1 = {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(1.1 * 2 * jnp.pi),
            'system_type': 'fluxonium',
            'shared_param_mark': 1,
            'arguments': {
                'phiext': 0.5 * 2 * jnp.pi
            }
        }
        params_2 = {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(1.2 * 2 * jnp.pi),
            'system_type': 'fluxonium',
            'shared_param_mark': 2,
            'arguments': {
                'phiext': 0.5 * 2 * jnp.pi
            }
        }
        params_3 = {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(1.0 * 2 * jnp.pi),
            'system_type': 'fluxonium',
            'shared_param_mark': 3,
            'arguments': {
                'phiext': 0.5 * 2 * jnp.pi
            }
        }
        params_4 = {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(0.8 * 2 * jnp.pi),
            'system_type': 'fluxonium',
            'shared_param_mark': 4,
            'arguments': {
                'phiext': 0.5 * 2 * jnp.pi
            }
        }
        params_5 = {
            'ec': jnp.array(1. * 2 * jnp.pi),
            'ej': jnp.array(4. * 2 * jnp.pi),
            'el': jnp.array(0.9 * 2 * jnp.pi),
            'system_type': 'fluxonium',
            'shared_param_mark': 5,
            'arguments': {
                'phiext': 0.5 * 2 * jnp.pi
            }
        }
        coupling_params = {
            'capacitive_coupling': {
                'strength': jnp.array(11.5e-3 * 2 * jnp.pi)
            },
            'inductive_coupling': {
                'strength': jnp.array(-1.0 * 2e-3 * 2 * jnp.pi)
            },
        }

        # adding attributes to nodes
        params = deque([params_1, params_2, params_3, params_4, params_5])
        for i in range(5):
            for j in range(5):
                temp_graph.nodes[(i, j)].update(params[j])
            # params list right shift
            params.rotate(3)
        # relabel nodes
        label_mapping = dict(
            (label,
             ''.join(['q', str(label[0]), str(label[1])]))
            for label in temp_graph.nodes)
        temp_graph = nx.relabel_nodes(temp_graph, label_mapping)
        # adding attributes to edges
        for edge in temp_graph.edges:
            temp_graph.edges[edge].update(coupling_params)
        # save temp_graph
        self.add_nodes_from(temp_graph.nodes.data())
        self.add_edges_from(temp_graph.edges.data())

        if seed is not None:
            # add variance to el ec ej params
            self.add_lcj_params_variance_to_graph(multi_err=0.01, seed=seed)


class XGatePeriodicGraph(PeriodicGraph):
    """Add pulse parameters for simultaneous X gates.
    In the example, we add 6 pulses on `q02`, `q03`, `q12`,
    `q13`, `q22` and `q23`.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)

        # note that the keywords of other pulse types are not always the same
        self.add_node(
            'q02', **{
                'pulse': {
                    'amp': jnp.array(0.07439612),
                    'length': jnp.array(40.05643032),
                    'omega_d': jnp.array(3.58624499),
                    'phase': jnp.array(1.29078953),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.33619935),
                    'post_comp': jnp.array(0.33863245)
                }
            })
        self.add_node(
            'q03', **{
                'pulse': {
                    'amp': jnp.array(0.06819317),
                    'length': jnp.array(40.04157743),
                    'omega_d': jnp.array(2.61055019),
                    'phase': jnp.array(0.5435847),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.51024),
                    'post_comp': jnp.array(-0.50827319)
                },
            })

        self.add_node(
            'q12', **{
                'pulse': {
                    'amp': jnp.array(0.07169389),
                    'length': jnp.array(40.0549766),
                    'omega_d': jnp.array(3.17153229),
                    'phase': jnp.array(0.87593417),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.0192981),
                    'post_comp': jnp.array(-0.00608968)
                }
            })
        self.add_node(
            'q13', **{
                'pulse': {
                    'amp': jnp.array(0.07708328),
                    'length': jnp.array(40.04376009),
                    'omega_d': jnp.array(4.19561078),
                    'phase': jnp.array(1.37478176),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.47187891),
                    'post_comp': jnp.array(0.47470512)
                }
            })

        self.add_node(
            'q22', **{
                'pulse': {
                    'amp': jnp.array(0.08007134),
                    'length': jnp.array(40.04477099),
                    'omega_d': jnp.array(5.00355723),
                    'phase': jnp.array(1.23550966),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.3009113),
                    'post_comp': jnp.array(0.30283622)
                }
            })
        self.add_node(
            'q23', **{
                'pulse': {
                    'amp': jnp.array(0.07334288),
                    'length': jnp.array(40.04449978),
                    'omega_d': jnp.array(3.51294492),
                    'phase': jnp.array(0.75852119),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.23767629),
                    'post_comp': jnp.array(-0.22839222)
                }
            })


class CNOTGatePeriodicGraph(PeriodicGraph):
    """Add pulse parameters for simultaneous CNOT gates.
    In the example, we add 3 CR pulses on `q02`, `q12` and `q22`.
    Note that the target qubits are `q03`, `q13`, `q23` respectively.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        # adding optimized pulse shape to edges
        self.add_node(
            'q02', **{
                'pulse': {
                    'amp': jnp.array(0.19362652),
                    'omega_d': jnp.array(2.63342338),
                    'phase': jnp.array(-0.56882493),
                    't_plateau': jnp.array(69.9963988),
                    't_ramp': jnp.array(30.20130596),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([0.26824387, -0.00582382, 0.19798702]),
                    'post_comp':
                        jnp.array([0.01907886, -0.0714353, -0.27239727])
                }
            })

        self.add_node(
            'q03', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-1.42126234, 0.68387943, 0.09867095]),
                    'post_comp':
                        jnp.array([-0.77910168, 0.16258883, -0.00519836])
                }
            })

        self.add_node(
            'q12', **{
                'pulse': {
                    'amp': jnp.array(0.18108519),
                    'omega_d': jnp.array(4.18157275),
                    'phase': jnp.array(0.27647924),
                    't_plateau': jnp.array(70.00154013),
                    't_ramp': jnp.array(29.98031495),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([0.00466943, 0.78741568, -0.0862072]),
                    'post_comp':
                        jnp.array([-0.00220794, 0.38970001, -1.51744927])
                }
            })

        self.add_node(
            'q13', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-1.09492337, -1.34049224, -1.11393925]),
                    'post_comp':
                        jnp.array([-0.40020693, 0.7668882, -0.74054753])
                },
            })

        self.add_node(
            'q22', **{
                'pulse': {
                    'amp': jnp.array(0.21301889),
                    'omega_d': jnp.array(3.51378339),
                    'phase': jnp.array(1.08374892),
                    't_plateau': jnp.array(69.94150468),
                    't_ramp': jnp.array(30.06869994),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([-0.00160062, -1.50823985, 0.16763715]),
                    'post_comp':
                        jnp.array(
                            [-1.03353877e-03, -1.75575364e+00, -1.10831887e-02])
                }
            })

        self.add_node(
            'q23', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-0.01605929, -0.1141277, -0.69661291]),
                    'post_comp':
                        jnp.array([1.99985832, -0.92168355, -0.49487715])
                },
            })


class CNOTGatePeriodicGraph2(PeriodicGraph):
    """Add pulse parameters for simultaneous CNOT gates.
    In the example, we add 3 CR pulses on `q02`, `q12` and `q22`.
    Note that the target qubits are `q01`, `q11`, `q21` respectively.
    The target qubits are different compared to CNOTGatePeriodicGraph.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        # adding optimized pulse shape to edges
        self.add_node(
            'q01', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-0.18729245, -0.26026559, 2.79556595]),
                    'post_comp':
                        jnp.array([-0.78930838, 1.19644611, -1.06109615]),
                },
            })

        self.add_node(
            'q02', **{
                'pulse': {
                    'amp': jnp.array(0.18788192),
                    'omega_d': jnp.array(4.607684),
                    'phase': jnp.array(-1.79225586),
                    't_plateau': jnp.array(69.95600166),
                    't_ramp': jnp.array(30.20569655),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array(
                            [-2.34791156e-04, 1.48276568e+00, 8.99221791e-01]),
                    'post_comp':
                        jnp.array([0.55455219, 0.00073962, 0.20388206]),
                }
            })

        self.add_node(
            'q11', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-0.09253156, 0.41565964, 0.26632383]),
                    'post_comp':
                        jnp.array([-0.75489047, 1.68428959, -1.71317095]),
                },
            })

        self.add_node(
            'q12', **{
                'pulse': {
                    'amp': jnp.array(0.10219415),
                    'omega_d': jnp.array(2.75130128),
                    'phase': jnp.array(-0.05679767),
                    't_plateau': jnp.array(69.25554591),
                    't_ramp': jnp.array(29.07256922),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([-0.0400265, -0.29959335, -3.0960532]),
                    'post_comp':
                        jnp.array([1.37150942, -0.0108739, -1.19826934]),
                }
            })

        self.add_node(
            'q21', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([0.50715299, -0.8000848, -0.72776202]),
                    'post_comp':
                        jnp.array([2.32881958, -0.63164007, -0.46285065]),
                },
            })

        self.add_node(
            'q22', **{
                'pulse': {
                    'amp': jnp.array(0.13038021),
                    'omega_d': jnp.array(4.19534038),
                    'phase': jnp.array(0.08719563),
                    't_plateau': jnp.array(70.01800142),
                    't_ramp': jnp.array(30.16955306),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array(
                            [-6.06081433e-04, -1.64773922e+00,
                             -2.04536211e+00]),
                    'post_comp':
                        jnp.array([3.13436934, 0.53835986, -0.79312629]),
                }
            })


class PeriodicGraphOpt(PeriodicGraph):
    """Device parameters for 5x5 periodic quantum processor graph after one-step
    optimization based on fidelities of simultaneous X gates and CNOT gates.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        # update device params by its gradient
        device_params = {
            'q02': {
                'ec': jnp.array(6.28318531),
                'ej': jnp.array(25.13274123),
                'el': jnp.array(6.28318531)
            },
            'q03': {
                'ec': jnp.array(6.28318531),
                'ej': jnp.array(25.13274123),
                'el': jnp.array(5.02654825)
            },
            'q12': {
                'ec': jnp.array(6.28318531),
                'ej': jnp.array(25.13274123),
                'el': jnp.array(5.65486678)
            },
            'q13': {
                'ec': jnp.array(6.28318531),
                'ej': jnp.array(25.13274123),
                'el': jnp.array(6.91150384)
            },
            'q22': {
                'ec': jnp.array(6.28318531),
                'ej': jnp.array(25.13274123),
                'el': jnp.array(7.53982237)
            }
        }
        device_grad = {
            'q02': {
                'ec': jnp.array(-7.00540723),
                'ej': jnp.array(2.24324625),
                'el': jnp.array(-6.1602346)
            },
            'q03': {
                'ec': jnp.array(2.81242553),
                'ej': jnp.array(-0.83988738),
                'el': jnp.array(2.5584191)
            },
            'q12': {
                'ec': jnp.array(5.98396226),
                'ej': jnp.array(-1.85285452),
                'el': jnp.array(5.35874343)
            },
            'q13': {
                'ec': jnp.array(-0.36019005),
                'ej': jnp.array(0.10861791),
                'el': jnp.array(-0.29491562)
            },
            'q22': {
                'ec': jnp.array(0.03386484),
                'ej': jnp.array(-0.02553209),
                'el': jnp.array(0.05359291)
            }
        }
        device_params = jax.tree_util.tree_map(lambda g, v: v - 0.01 * g,
                                               device_grad, device_params)
        device_params.update({
            'capacitive_coupling_all_unify': {
                'strength': jnp.array(0.07225663)
            },
            'inductive_coupling_all_unify': {
                'strength': jnp.array(-0.01256637)
            },
        })
        self.update_params(device_params, unify_coupling=True)


class XGatePeriodicGraphOpt(PeriodicGraphOpt):
    """Add pulse parameters for simultaneous X gates.
    In the example, we add 6 pulses on `q02`, `q03`, `q12`,
    `q13`, `q22` and `q23`.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        # adding optimized pulse shape to edges
        self.add_node(
            'q02', **{
                'pulse': {
                    'amp': jnp.array(0.07467672),
                    'length': jnp.array(39.82459069),
                    'omega_d': jnp.array(3.70771137),
                    'phase': jnp.array(0.7558433),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.42591073),
                    'post_comp': jnp.array(-0.41941967)
                }
            })
        self.add_node(
            'q03', **{
                'pulse': {
                    'amp': jnp.array(0.06839676),
                    'length': jnp.array(39.81302627),
                    'omega_d': jnp.array(2.56897537),
                    'phase': jnp.array(0.85390569),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.11757419),
                    'post_comp': jnp.array(-0.11531302)
                }
            })

        self.add_node(
            'q12', **{
                'pulse': {
                    'amp': jnp.array(0.07142831),
                    'length': jnp.array(39.85070846),
                    'omega_d': jnp.array(3.07586917),
                    'phase': jnp.array(1.58941357),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.74646573),
                    'post_comp': jnp.array(0.74593951)
                }
            })
        self.add_node(
            'q13', **{
                'pulse': {
                    'amp': jnp.array(0.07763859),
                    'length': jnp.array(39.85217838),
                    'omega_d': jnp.array(4.20299531),
                    'phase': jnp.array(1.43135504),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.58423741),
                    'post_comp': jnp.array(0.58729911)
                }
            })

        self.add_node(
            'q22', **{
                'pulse': {
                    'amp': jnp.array(0.08054434),
                    'length': jnp.array(39.80346128),
                    'omega_d': jnp.array(5.00262381),
                    'phase': jnp.array(1.39100452),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(-0.48806854),
                    'post_comp': jnp.array(0.49022838)
                }
            })
        self.add_node(
            'q23', **{
                'pulse': {
                    'amp': jnp.array(0.07422721),
                    'length': jnp.array(39.85217838),
                    'omega_d': jnp.array(3.63596315),
                    'phase': jnp.array(0.26554983),
                    'pulse_type': 'cos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(1.03746918),
                    'post_comp': jnp.array(-1.02909962)
                }
            })


class CNOTGatePeriodicGraphOpt(PeriodicGraphOpt):
    """Add pulse parameters for simultaneous CNOT gates.
    In the example, we add 3 CR pulses on `q02`, `q12`
    and `q22`.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        # adding optimized pulse shape to edges
        self.add_node(
            'q02', **{
                'pulse': {
                    'amp': jnp.array(0.18128846),
                    'omega_d': jnp.array(2.58934559),
                    'phase': jnp.array(-0.24290228),
                    't_plateau': jnp.array(69.93608145),
                    't_ramp': jnp.array(29.92806488),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([0.05671005, 0.02682845, 0.0395099]),
                    'post_comp':
                        jnp.array([-0.01528741, 0.04418319, -0.374539]),
                }
            })

        self.add_node(
            'q03', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-1.46070525, 0.66029355, 0.06603344]),
                    'post_comp':
                        jnp.array([-1.00629771, 0.16411027, -0.00797514]),
                },
            })

        self.add_node(
            'q12', **{
                'pulse': {
                    'amp': jnp.array(0.17872194),
                    'omega_d': jnp.array(4.19989714),
                    'phase': jnp.array(-0.01543561),
                    't_plateau': jnp.array(69.95327862),
                    't_ramp': jnp.array(29.94562879),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array([0.00185588, 1.00099539, -0.22439037]),
                    'post_comp':
                        jnp.array([0.00200104, 0.58679953, -1.74804011]),
                }
            })

        self.add_node(
            'q13', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([-1.02837338, -1.10918103, -0.80985523]),
                    'post_comp':
                        jnp.array([-0.11690826, 1.09973631, -1.0941268]),
                },
            })

        self.add_node(
            'q22', **{
                'pulse': {
                    'amp': jnp.array(0.23370657),
                    'omega_d': jnp.array(3.65018191),
                    'phase': jnp.array(-0.50006247),
                    't_plateau': jnp.array(69.96302405),
                    't_ramp': jnp.array(29.98769304),
                    'pulse_type': 'rampcos',
                    'operator_type': 'phi_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp':
                        jnp.array(
                            [-1.90040736e-04, -1.34752499e+00, 3.15152817e-01]),
                    'post_comp':
                        jnp.array(
                            [-1.42308406e-03, -1.59853235e+00,
                             -1.88352743e-01]),
                }
            })

        self.add_node(
            'q23', **{
                'compensation': {
                    'pre_comp':
                        jnp.array([0.36516325, -0.38763126, -0.41575102]),
                    'post_comp':
                        jnp.array([1.65404972, -0.67251125, -0.04376101]),
                },
            })


"""
// JSON schema for nodes

{
  "$schema": "http://json-schema.org/draft-07/schema#",
  "type": "object",
  "properties": {
    "ec": {
      "type": "array",
      "items": {
        "description": "charging energy in unit GHz.
                We multiply 2pi to them so that we follow the same convention of
                setting hbar=1 as done in qutip's sesolve.
                Same for ej and el",
        "type": "number"
      }
    },
    "ej": {
      "type": "array",
      "items": {
        "description": "Josephson energy in unit GHz",
        "type": "number"
      }
    },
    "el": {
      "type": "array",
      "items": {
        "description": "inductive energy in unit GHz",
        "type": "number"
      }
    },
    "system_type": {
      "type": "string",
      "enum": [
        "fluxonium",
        "transmon",
        "resonator"
      ]
    },
    "shared_param_mark": {
      "type": "integer"
      "description": "We will set the parameters to be the same for
                    fluxoniums with same shared_param_mark.
                    This is intended to only effect gradient computation when share_params = True in Evolve,
                    (see hamiltonian/common_functions/compute_unitary_evolve.py)
                    but to be safe just make sure to set shared fluxoniums to the same value.""
    },
    "arguments": {
      "type": "object",
      "properties": {
        "phiext": {
          "description": "phiext is the value of phi_ext used in the fluxonium,
                            e.g. phiext = 0.5 * 2 * pi means that the fluxonium is at the sweetspot",
          "type": "number"
        },
        "phi_max": {
          "description": "phi basis range [-phi_max, phi_max), must be multiple of pi",
          "type": "number"
        },
        "truncated_dim": {
          "description": "How many eigenvectors we keep for each subsystem.
                            Currently we always keep the eigenvectors with lowest k energies",
          "type": "number"
        }
      }
    },
    "variance": {
      "type": "object",
      "description": "The variance of ec, ej, el parameters.
                        The values here are multiplied to the ec ej el if add_random = True in Evolve
                        see hamiltonian/common_functions/compute_unitary_evolve.py",
      "properties": {
        "ec": {
          "type": "array",
          "items": {
            "type": "number"
          }
        },
        "ej": {
          "type": "array",
          "items": {
            "type": "number"
          }
        },
        "el": {
          "type": "array",
          "items": {
            "type": "number"
          }
        }
      }
    },
    "pulse": {
      "type": "object",
      "description": "the parameters below depend on the pulse type",
      "properties": {
        "amp": {
          "type": "array",
          "items": {
            "description": "the amplitude of pulse",
            "type": "number"
          }
        },
        "length": {
          "type": "array",
          "items": {
            "description": "the length of pulse",
            "type": "number"
          }
        },
        "omega_d": {
          "type": "array",
          "items": {
            "description": "drive frequency of modulate wave",
            "type": "number"
          }
        },
        "phase": {
          "type": "array",
          "items": {
            "description": "phase of modulate wave",
            "type": "number"
          }
        },
        "pulse_type": {
          "type": "string",
          “description": "see hamiltonian/time_evolution/pulseshape.py"
          "enum": [
            "trapezoid",
            "cos",
            "rampcos",
            "tanh",
            "erf",
            "gaussian"
          ]
        },
        "operator_type": {
          "type": "string",
          "enum": [
            "n_operator",
            "phi_operator"
          ]
        },
        "delay": {
          "description": "time delay for waiting other gate operations",
          "type": "number"
        },
        "crosstalk": {
          "description": "the qubit has crosstalk",
          "type": "object",
          "additionalProperties": {
            "description": "the factor of crosstalk amplitude",
            "type": "number"
          }
        },
        "arguments": {
          "type": "object",
          "properties": {
            "modulate_wave": {
              "type": "boolean"
              "description": "if true, a modulate wave of frequency omega_d is
                                multiplied to the waveform"
            }
          }
        }
      }
    }
  }
}

// JSON schema for edges

{
  "$schema": "http://json-schema.org/draft-07/schema#",
  "type": "object",
  "properties": {
    "capacitive_coupling": {
      "type": "object",
      "properties": {
        "strength": {
          "description": "coefficient for the capacitive interaction strength",
          "type": "array",
          "items": {
            "type": "number"
          }
        }
      }
    },
    "inductive_coupling": {
      "type": "object",
      "properties": {
        "strength": {
          "description": "coefficient for the inductive interaction strength",
          "type": "array",
          "items": {
            "type": "number"
          }
        }
      }
    }
  }
}

"""



import os
import sys
import jax
import jax.numpy as jnp
import pytest

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append('/'.join(dir_path.split('/')[:-1]))

from benchmark.utils.create_simultaneous_model import (create_simultaneous_x,
                                                       create_simultaneous_cnot)

trotter_order_list = [2, 4j, 4, None]
diag_ops_list = [True, False]


@pytest.mark.parametrize('trotter_order', trotter_order_list)
@pytest.mark.parametrize('diag_ops', diag_ops_list)
def test_simultaneous_x_gradient(trotter_order, diag_ops):
    n_qubit = 2
    astep = 3000
    ref_evo = create_simultaneous_x(n_qubit=n_qubit,
                                    astep=astep,
                                    trotter_order=trotter_order,
                                    diag_ops=diag_ops)

    @jax.grad
    def cal_ref_grad(params):
        u_cal = ref_evo.product_basis(params)
        return jnp.sum(u_cal).real

    ref_grad = cal_ref_grad(ref_evo.all_params)

    evo = create_simultaneous_x(n_qubit=n_qubit,
                                astep=astep,
                                trotter_order=trotter_order,
                                diag_ops=diag_ops,
                                custom_vjp=True)

    @jax.grad
    def cal_grad(params):
        u_cal = evo.product_basis(params)
        return jnp.sum(u_cal).real

    grad = cal_grad(evo.all_params)

    assert all(
        jax.tree.leaves(
            jax.tree_util.tree_map(lambda a, b: jnp.allclose(a, b), ref_grad,
                                   grad)))


@pytest.mark.slow
@pytest.mark.parametrize('trotter_order', trotter_order_list)
@pytest.mark.parametrize('diag_ops', diag_ops_list)
def test_simultaneous_cnot_gradient(trotter_order, diag_ops):
    n_qubit = 2
    astep = 5000
    ref_evo = create_simultaneous_cnot(n_qubit=n_qubit,
                                       astep=astep,
                                       trotter_order=trotter_order,
                                       diag_ops=diag_ops)

    @jax.grad
    def cal_ref_grad(params):
        u_cal = ref_evo.product_basis(params)
        return jnp.sum(u_cal).real

    ref_grad = cal_ref_grad(ref_evo.all_params)

    evo = create_simultaneous_cnot(n_qubit=n_qubit,
                                   astep=astep,
                                   trotter_order=trotter_order,
                                   diag_ops=diag_ops,
                                   custom_vjp=True)

    @jax.grad
    def cal_grad(params):
        u_cal = evo.product_basis(params)
        return jnp.sum(u_cal).real

    grad = cal_grad(evo.all_params)

    assert all(
        jax.tree.leaves(
            jax.tree_util.tree_map(lambda a, b: jnp.allclose(a, b, rtol=1e-4),
                                   ref_grad, grad)))



from abc import abstractmethod
from typing import Optional
import numpy as np
import jax.numpy as jnp

from .base import QuantumSystem


class CircuitLCJ(QuantumSystem):
    """Class for a generic L-C-J circuit.
    """

    def __init__(self,
                 b_transmon: bool = True,
                 basis: str = "phase",
                 num_basis: int = 0,
                 truncated_dim: int = 0,
                 n_max: int = 0,
                 phi_max: float = 0,
                 is_basis_sym: bool = False,
                 drive_for_state_phase: str = "charge",
                 name: str = 'circuit_lcj'):
        QuantumSystem.__init__(self, name=name)
        # Save basis related parameters
        self.basis = basis

        # Initialize basis related variables
        self.phi_max_int: int = 0  # phi_max / pi, must be integer
        self.n_max: int = 0  # n_max, must be integer
        self.phi_max: float = 0  # phi_max

        # Setup basis set
        self.b_transmon = b_transmon

        # Deal with number of basis
        self.is_basis_sym = is_basis_sym
        if is_basis_sym:
            if num_basis % 2 == 0:
                num_basis += 1

        # Sync parameter between two basis sets
        # To work with two basis sets together,
        # the n range  * phi range must be 2pi * points
        if self.basis in ("phase", "charge", "phase_only"):
            # n_max is always used first and override others if necessary
            if n_max == 0:
                # Check if phi_max is valid
                phi_max_int = phi_max / np.pi
                if np.abs(np.rint(phi_max_int) - phi_max_int) > 1e-4:
                    raise ValueError("phi_max must be a multiple of pi")
                phi_max_int = int(np.rint(phi_max_int))
                n_max = int(np.rint(2 * np.pi * num_basis / phi_max / 2) // 2)
            elif phi_max == 0:
                phi_max_int = num_basis // (n_max * 2)
                phi_max = phi_max_int * np.pi

            # Check the number of basis
            num_basis_new = int(np.rint(2 * n_max * phi_max / np.pi))
            if num_basis != 0:
                if num_basis != num_basis_new:
                    print("num_basis changed from %i to %i"
                          " to be a multiple of max range" %
                          (num_basis, num_basis_new))
                num_basis = num_basis_new
            # Check consistency
            # All information exists, check consistency
            if abs(2 * n_max * phi_max - np.pi * num_basis) > 1e-4:
                print(2 * n_max * phi_max / np.pi, num_basis)
                raise ValueError("phi_max, n_max, num_basis are inconsistent")

            # Ensure this is set
            phi_max_int = int(np.rint(phi_max / np.pi))

            self.phi_max = phi_max
            self.phi_max_int = phi_max_int
            self.n_max = n_max

        self.num_basis = num_basis
        self.truncated_dim = truncated_dim
        # print(self.phi_max, self.phi_max_int, self.n_max, self.num_basis)
        # Define phase basis
        self.num_phi = None  # number of phi basis (should be same as num_basis)
        self.ar_phi = None  # array of phi values
        self.delta_phi = None  # Difference between two phi basis
        # Define charge basis
        self.ar_n = None
        self.charge_cos = None
        self.charge_sin = None
        # Define the phase-charge conversion
        self.u1_n_phi = None
        self.u2_n_phi = None

        if self.basis in ("phase", "charge", "phase_only"):
            self.set_phi_basis(phi_max=phi_max, num_phi=num_basis)
        if self.basis in ("phase", "charge"):
            self.set_charge_basis(n_max=n_max, num_n=num_basis)
        if self.basis in ("phase", "charge"):
            self.set_n_phi_transform()

        # fix phase
        self.drive_for_state_phase = drive_for_state_phase
        # cache
        self.eigvec_in_raw_basis = None

    def set_phi_basis(self,
                      phi_max: float,
                      num_phi: Optional[int] = None,
                      phi_step: Optional[float] = None):
        """Initializes parameters for phase basis.

        One may choose to initialize with number of phi values or phi step size.

        Args:
            phi_max : [-phi_max, phi_max-1] as basis
            num_phi : Number of points between phi_max.
            phi_step : Step size of phi.
        """
        self.phi_max = phi_max
        if num_phi is not None:
            self.num_phi = num_phi  # Number of discretized num_phi basis
        elif phi_step is not None:
            self.num_phi = int(phi_max * 2 / phi_step)
            if not self.b_transmon:
                self.num_phi += 1

        # is_basis_even control if end point is included or not
        self.ar_phi = np.linspace(-self.phi_max,
                                  self.phi_max,
                                  self.num_basis,
                                  endpoint=self.is_basis_sym)

        self.delta_phi = self.ar_phi[1] - self.ar_phi[0]

    def set_charge_basis(self, n_max, num_n=0):
        """Initializes parameters for charge basis.

        cos/sin works as exp(ipt) which increase/decrease n by 1

        Note num_n will not be exact as x+any integer belongs to one group for cos/sin
        so a multiple of 2*n_max is used

        Args:
            n_max : Basis as [-n_max, n_max-1]
                (or n_max depends on *is_basis_sym*)
            num_n : Number of basis, not used for transmon
                (where n basis is periodic and must be integer,
                so fully determined by *n_max*)
        """
        n_set = self.phi_max_int  # Number of points per integer is the same as phi_max/pi
        n_group = n_max * 2 + (1 if self.is_basis_sym else 0)
        n = n_set * n_group
        self.ar_n = np.linspace(-n_max, n_max, n, endpoint=self.is_basis_sym)

        mcos = np.zeros((n, n), dtype=np.complex128)
        msin = np.zeros((n, n), dtype=np.complex128)
        # Add block-diaogonal matrix element
        # (only values differ by 1 are connected)
        for i1 in range(n_group - 1):
            i2 = i1 + 1
            for j in range(n_set):
                mcos[i1 * n_set + j, i2 * n_set + j] = 1
                mcos[i2 * n_set + j, i1 * n_set + j] = 1
                msin[i1 * n_set + j, i2 * n_set + j] = -1 / 1j
                msin[i2 * n_set + j, i1 * n_set + j] = 1 / 1j
        self.charge_cos = mcos / 2
        self.charge_sin = msin / 2

    def create_d2phi(self) -> np.ndarray:
        r"""Computes :math:`\frac{d^2}{d \phi^2}`

        This is from finite difference, which is an approximation.
        Used only in **phase_only** basis.

        Returns:
            matrix of :math:`\frac{d^2}{d \phi^2}`
        """
        m = np.eye(self.num_phi, dtype=np.complex128) * -2
        for i in range(0, self.num_phi):
            if i - 1 >= 0:
                m[i, i - 1] = 1
            if i + 1 < self.num_phi:
                m[i, i + 1] = 1

        # Periodic boundary condition
        if self.b_transmon:
            m[0, -1] = 1
            m[-1, 0] = 1

        return m / self.delta_phi**2

    def create_dphi(self) -> np.ndarray:
        r"""Computes :math:`\frac{d}{d \phi}`

        This is from finite difference, which is an approximation.
        Used only in **phase_only** basis.

        Returns:
            matrix of :math:`\frac{d}{d \phi}`
        """
        m = np.zeros((self.num_phi, self.num_phi), dtype=np.complex128)
        for i in range(0, self.num_phi):
            if i - 1 >= 0:
                m[i, i - 1] = -1
            if i + 1 < self.num_phi:
                m[i, i + 1] = 1

        # Periodic boundary condition
        if self.b_transmon:
            m[0, -1] = -1
            m[-1, 0] = 1

        return m / (2 * self.delta_phi)

    def create_phi(self) -> np.ndarray:
        """Computes the phi matrix.

        Returns:
            matrix of phase
        """
        if self.basis == "phase" or self.basis == "phase_only":
            return np.diag(self.ar_phi)
        elif self.basis == "charge":
            return self.transform_phi_to_n(np.diag(self.ar_phi))

    def create_n(self):
        r"""Computes charge matrix.

        Follows :math:`n = -i\frac{\partial}{\partial \phi}`

        For **phase_only** basis, use center difference,
        :math:`\frac{d}{d\phi}(N) = \phi(N+1)-\phi(N-1)` for phase basis.
        This is not accurate and this basis set should not be used.

        Use charge -> phase basis transformation for better accuracy,
        which is in **phase** and **charge** basis.

        Returns:
            matrix of charge
        """
        if self.basis == "phase_only":
            m = self.create_dphi()
            m = -1j * m
            return m
        elif self.basis == "charge":
            m = np.diag(self.ar_n.copy())
            return m
        elif self.basis == "phase":
            m = np.diag(self.ar_n.copy())
            return self.transform_n_to_phi(m)

    def set_n_phi_transform(self):
        """Computes the unitary transformation between n and phi.

        Store as u_n_phi
        phi within -pi, pi and n is bounded to n_max

        Note this transformation is ill defined if number of basis are different

        Require different normalization factor from phi->n and n->phi

        Require both set_charge_basis and set_phase_basis are done
        """
        x, y = np.meshgrid(self.ar_n, self.ar_phi)
        self.u1_n_phi = np.exp(1j * x * y)
        self.u2_n_phi = np.conj(self.u1_n_phi.T)

    def transform_n_to_phi(self, mat: np.ndarray) -> np.ndarray:
        """Run change of basis of matrix from charge basis to phase basis

        Args:
            mat: the matrix in charge basis

        Returns:
            the matrix in phase basis
        """
        return (self.u1_n_phi @ mat @ self.u2_n_phi) / self.ar_phi.size

    def transform_phi_to_n(self, mat: np.ndarray) -> np.ndarray:
        """Run change of basis of matrix from phi basis to n basis

        Args:
            mat: the matrix in phase basis

        Returns:
            the matrix in charge basis
        """
        return (self.u2_n_phi @ mat @ self.u1_n_phi) / self.ar_n.size

    @property
    def qdevice_type(self):
        return 'qubit'

    @property
    def dim(self) -> int:
        """Returns truncated Hilbert space dimension"""
        return self.truncated_dim

    @abstractmethod
    def idling_hamiltonian(self) -> jnp.ndarray:
        """Create the Hamiltonian matrix of the qubit.

        Returns:
            ndarray of the Hamiltonian matrix
        """

    def eigenenergies(self, unify_state_phase=True):
        """Returns array of eigenvalues.

        Args:
            evals_count:
                number of desired eigenvalues (default value = 6)
            unify_state_phase(bool): if True, adjusts the phase of eigenstates
                to meet the condition <i|m|i+1> matrix elements are real positive.
        """
        self.eigval, self.eigvec_in_raw_basis = self._calc_eigsys()
        if unify_state_phase:
            # Convert the basis to let n be real positive in 0-1 1-2 ...
            if self.drive_for_state_phase == "charge":
                opt = self.create_n()
            elif self.drive_for_state_phase == "phase":
                opt = self.create_phi()
            opt_eigenbasis = jnp.conj(
                self.eigvec_in_raw_basis).T @ opt @ self.eigvec_in_raw_basis
            self.unify_state_phase(opt_eigenbasis)

        return self.eigval[:self.dim]

    def unify_state_phase(self, operator: jnp.ndarray):
        """Adjusts the phase of eigenstates to meet the condition <i|m|i+1>
        matrix elements are real positive.

        This is to follow the simple idea we
        are always looking at phase defined by pulse.
        So for charge drive, let the <0|n|1>  be real positive is natural.
        For flux drive similar condition exists.

        Only bands within truncated_dim are considered.

        Require eigvec_in_raw_basis computed before calling this function.

        Args:
            operator: This is to specify which drive is used, so
                eigenstates are rotated to let <i|drive|i+1> be real positive
                Possible values are ``n_operator`` (in eigenbasis) and
                ``phi_operator`` (in eigenbasis).

        Returns:
            Modified eigvec_in_raw_basis.
        """
        if self.eigvec_in_raw_basis is None:
            raise ValueError("Must compute eigvec_in_raw_basis before rotating")

        ar_angle = jnp.cumsum(jnp.angle(jnp.diag(operator, k=1)))
        ar_angle = jnp.concatenate([jnp.zeros(1), ar_angle])
        # Apply the angle to basis
        ar_phase_eigenbasis = jnp.exp(ar_angle * -1j)

        # numpy broadcast
        self.eigvec_in_raw_basis = self.eigvec_in_raw_basis * ar_phase_eigenbasis

        return self.eigvec_in_raw_basis



import os
import numpy as np

import supergrad
from supergrad.quantum_system import Fluxonium, Transmon

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)


def test_transmon_basis():
    # construct model
    class ExploreTransmon(supergrad.Helper):

        def __init__(self, basis) -> None:
            super().__init__(basis=basis)

        def _init_quantum_system(self):
            self.tmon = Transmon(ec=1.2,
                                 ej=30.02,
                                 ng=0.3,
                                 phiext=0.,
                                 basis=self.kwargs['basis'],
                                 truncated_dim=6,
                                 drive_for_state_phase='charge',
                                 n_max=31,
                                 constant=True)

        def energy_spectrum(self):
            return self.tmon.eigenenergies()

        def operator(self):
            """Calculate the operator in the raw basis."""
            return np.asarray(
                [self.tmon.n_operator(),
                 self.tmon.phi_operator()])

    tmon_phase = ExploreTransmon('phase')
    tmon_charge = ExploreTransmon('charge')
    # compare spectra
    assert np.allclose(tmon_charge.energy_spectrum({}),
                       tmon_phase.energy_spectrum({}))
    # compare operator
    assert np.allclose(tmon_charge.operator({}), tmon_phase.operator({}))


def test_tunable_transmon_basis():
    # construct model
    class ExploreTransmon(supergrad.Helper):

        def __init__(self, basis) -> None:
            super().__init__(basis=basis)

        def _init_quantum_system(self):
            self.tmon = Transmon(ec=0.5,
                                 ej=50.0,
                                 d=0.01,
                                 phiext=0.1 * 2 * np.pi,
                                 ng=0.3,
                                 basis=self.kwargs['basis'],
                                 truncated_dim=6,
                                 drive_for_state_phase='charge',
                                 n_max=31,
                                 constant=True)

        def energy_spectrum(self):
            return self.tmon.eigenenergies()

        def operator(self):
            """Calculate the operator in the raw basis."""
            return np.asarray(
                [self.tmon.n_operator(),
                 self.tmon.phi_operator()])

    tmon_phase = ExploreTransmon('phase')
    tmon_charge = ExploreTransmon('charge')
    # compare spectra
    assert np.allclose(tmon_charge.energy_spectrum({}),
                       tmon_phase.energy_spectrum({}))
    # compare operator
    assert np.allclose(tmon_charge.operator({}), tmon_phase.operator({}))


def test_fluxonium_basis():
    # construct model
    class ExploreFluxonium(supergrad.Helper):

        def __init__(self, basis) -> None:
            super().__init__(basis=basis)

        def _init_quantum_system(self):
            self.fmon = Fluxonium(ec=2.5,
                                  ej=8.9,
                                  el=0.5,
                                  phiext=0.33 * 2 * np.pi,
                                  basis=self.kwargs['basis'],
                                  truncated_dim=6,
                                  drive_for_state_phase='charge',
                                  put_phiext_on_inductor=True,
                                  phi_max=5 * np.pi,
                                  constant=True)

        def energy_spectrum(self):
            return self.fmon.eigenenergies()

        def operator(self):
            """Calculate the operator in the raw basis."""
            return np.asarray(
                [self.fmon.n_operator(),
                 self.fmon.phi_operator()])

    fmon_phase = ExploreFluxonium('phase')
    fmon_charge = ExploreFluxonium('charge')
    # compare spectra
    assert np.allclose(fmon_charge.energy_spectrum({}),
                       fmon_phase.energy_spectrum({}))
    # compare operator
    assert np.allclose(fmon_charge.operator({}), fmon_phase.operator({}))


# @pytest.mark.slow
# def test_hermite_basis():
#     # construct model
#     class ExploreFluxonium(supergrad.Helper):

#         def __init__(self, basis) -> None:
#             super().__init__(basis=basis)

#         def _init_quantum_system(self):
#             self.fmon = Fluxonium(ec=2.5,
#                                   ej=8.9,
#                                   el=0.5,
#                                   phiext=0.33 * 2 * np.pi,
#                                   basis=self.kwargs['basis'],
#                                   truncated_dim=6,
#                                   drive_for_state_phase='charge',
#                                   put_phiext_on_inductor=False,
#                                   phi_max=5 * np.pi,
#                                   constant=True)

#         def energy_spectrum(self):
#             return self.fmon.eigenenergies()

#         def operator(self):
#             """Calculate the operator in the raw basis."""
#             return np.asarray(
#                 [self.fmon.n_operator(),
#                  self.fmon.phi_operator()])

#     fmon_phase = ExploreFluxonium('phase')
#     fmon_hermite = ExploreFluxonium('hermite')
#     print(
#         np.max(
#             np.abs(
#                 fmon_hermite.energy_spectrum({}) -
#                 fmon_phase.energy_spectrum({}))))
#     # compare spectra
#     assert np.allclose(fmon_hermite.energy_spectrum({}),
#                        fmon_phase.energy_spectrum({}))
#     # compare operator
#     assert np.allclose(fmon_hermite.operator({}), fmon_phase.operator({}))



from functools import reduce
import numpy as np
import jax.numpy as jnp
from jax.scipy.stats import cauchy
import haiku as hk
import matplotlib.pyplot as plt

from supergrad.quantum_system import (Fluxonium, Resonator, InteractingSystem,
                                      parse_interaction)


def wrapper_yaqcs_data(data):
    """A wrapper for data from yaqcs experiment.

    Args:
        data: `DataSet` from yaqcs
    """
    # Return to base unit
    data.remove_units()
    # Format data
    if not data.is_adaptive:
        regular_grid = [True] * len(data.axis_names)
    else:
        regular_grid = data['regular_grid']
    # Get the result of experiment in complex number
    if not all(regular_grid):
        data = data.to_nested_list()
        assert 'mag' in data and 'phase' in data
        exp_array = []
        for mag, phase in zip(data['mag'], data['phase']):
            exp_array.append(10**(mag / 10) * np.exp(1j * phase))
        grid = []
        for ii, regular in enumerate(regular_grid):
            if regular:
                grid.append(data[data.axis_names[ii]])
            else:
                # unify non-regular grid
                name: str = data.axis_names[ii]
                nrgrid = data[name]
                union_nrgrid = reduce(np.union1d, nrgrid)
                grid.append(union_nrgrid)
                # full the union grid
                new_exp_array = []
                for jj, exp in enumerate(exp_array):
                    mask = np.isin(union_nrgrid, nrgrid[jj])
                    template = np.full(union_nrgrid.shape,
                                       np.nan,
                                       dtype=complex)
                    template[mask] = exp
                    new_exp_array.append(template)
        exp_array = np.stack(new_exp_array)

    else:
        data.compute_complex_from_mag_phase()
        exp_array = data['complex']
        grid = []
        for name in data.axis_names:
            grid.append(data[name])

    return tuple(grid + [
        exp_array,
    ])


def wrapper_phi_freq(bias_list, freq_list, periodic=True):
    """A wrapper that format `bias_list` and `freq_list`.

    Args:
        bias_list: the bias to adjust phi external in Fluxonium.
        freq_list: the frequency list.
        periodic: True for using periodic boundary condition of phi in simulation.
    """
    bias_list = jnp.asarray(bias_list)
    freq_list = jnp.asarray(freq_list) / 1e9  # unit in GHz
    # convert bias to phi_ext/phi_0
    bias_phi0 = hk.get_parameter('bias_phi0', [], init=jnp.zeros)
    bias_phipi = hk.get_parameter('bias_phipi', [], init=jnp.zeros)

    phi_list = (bias_list - bias_phi0) / (bias_phipi - bias_phi0) / 2
    if periodic:
        # use periodic boundary condition of phi
        mod_phi = jnp.modf(2 * jnp.modf(phi_list)[0])
        phi_list = (mod_phi[0] - mod_phi[1]) / 2

    return phi_list, freq_list


def truncated_mse(a, b, trancated_percentile: int = 95):
    """Truncated mean squared error"""

    a = jnp.asarray(a)
    b = jnp.asarray(b)
    mse = (a - b)**2
    # truncated differences
    mse = jnp.minimum(mse, jnp.percentile(mse, trancated_percentile))
    return jnp.sum(mse)


def _cal_spectrum_vs_phi(phi, phi_max=5 * np.pi, num_basis=50, **kwargs):
    # initial quantum system
    fq = Fluxonium(phiext=0.0, phi_max=phi_max, num_basis=num_basis, **kwargs)

    # calculate spectrum
    def calc_spectrum_vs_phi(phi):
        fq.phiext = phi * 2 * np.pi
        return fq.eigenenergies()

    # vmap it
    if len(phi) > 1:
        return hk.vmap(calc_spectrum_vs_phi, split_rng=False)(phi)
    else:
        return calc_spectrum_vs_phi(phi)


def _cal_transmission_vs_phi(phi,
                             num_photon,
                             eval_count=3,
                             phi_max=5 * np.pi,
                             num_basis=50,
                             truncated_dim=12):

    def calc_transmission_vs_phi(phi):

        # Physical components
        fqubit = Fluxonium(phiext=0.0,
                           phi_max=phi_max,
                           num_basis=num_basis,
                           truncated_dim=truncated_dim)
        cavity = Resonator(truncated_dim=12)
        subsystem_list = [fqubit, cavity]
        # specifying interactions between subsystems
        # There are some operators(`Callable` function`) in the components.
        inter_term = parse_interaction(op0=fqubit.n_operator,
                                       op1=cavity.annihilation_operator,
                                       add_hc=True)
        # the option `add_hc=True` adds the hermitian conjugate to the Hamiltonian
        # now instance InteractingSystem to handle components and InteractionTerms
        # one can set module name here
        hilbertspace = InteractingSystem(subsystem_list,
                                         inter_term,
                                         name='fluxonium_cavity')
        # override phiext
        hilbertspace['fluxonium'].phiext = phi * 2 * np.pi
        energy_nd = hilbertspace.compute_energy_map()
        return energy_nd[:eval_count, num_photon] - energy_nd[:eval_count,
                                                              num_photon - 1]

    # vmap it
    if len(phi) > 1:
        return hk.vmap(calc_transmission_vs_phi, split_rng=False)(phi)
    else:
        return calc_transmission_vs_phi(phi)


@hk.without_apply_rng
@hk.transform
def loss_fit_spectrum(bias_list, freq_list, end_level, periodic=True):
    """Fits parameters from fluxonium spectrum vs bias data.

    Args:
        bias_list (list[~numpy.ndarray]): Bias data.
        freq_list (list[~numpy.ndarray]): Spectrum data.
        end_level (int): Ending level for spectrum.
        periodic (bool): Using periodic boundary condition of phi
            when the target phi is beyond the phi basis range.
    """
    phi_list, freq_list = wrapper_phi_freq(bias_list, freq_list, periodic)

    spec_out = _cal_spectrum_vs_phi(phi_list)
    sim_freq_list = spec_out[:, end_level] - spec_out[:, 0]

    return truncated_mse(sim_freq_list, freq_list, 100)


@hk.without_apply_rng
@hk.transform
def loss_kl_div(xs, ys, exp_array, plot=False):
    """Loss function for fit qubit spectra by the Kullback-Leibler divergence.

    Args:
        data: dateset from experiment.
    """
    # Normalize
    data_exp = jnp.abs(exp_array -
                       jnp.nanmean(exp_array, axis=1, keepdims=True))
    data_exp = data_exp / jnp.nansum(data_exp)
    data_exp = jnp.nan_to_num(data_exp, nan=1e-4)
    if plot:
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].pcolormesh(xs, ys, data_exp.T)
        axs[0].set_title(r'P(x, $\varepsilon$)')
        axs[0].set_xlabel(r'x')
        axs[0].set_ylabel(r'$\varepsilon [\text{GHz}] $')

    # convert bias to phi_ext/phi_0
    phi_list, ys = wrapper_phi_freq(xs, ys)

    spec_out = _cal_spectrum_vs_phi(phi_list)
    peak_list = spec_out[:, 1] - spec_out[:, 0]

    # set a bandwidth
    bandwidth = hk.get_parameter('bandwidth', [], init=jnp.zeros)
    noise = hk.get_parameter('noise', [], init=jnp.zeros)
    sim_data = []
    for peak in peak_list:
        sim_data.append(cauchy.pdf(ys, loc=peak, scale=bandwidth))
    sim_data = jnp.asarray(sim_data)

    # add noise
    sim_data += noise
    sim_data = sim_data / jnp.sum(sim_data)

    # plot
    if plot:
        axs[1].pcolormesh(xs, ys, sim_data.T)
        axs[1].set_title(r'Q(x, $\varepsilon$)')
        axs[1].set_xlabel(r'x')
        axs[1].set_ylabel(r'$\varepsilon [\text{GHz}] $')
        plt.show()
    # calculate kl div
    return jnp.nansum(data_exp * jnp.log(data_exp / sim_data))


@hk.without_apply_rng
@hk.transform
def loss_kl_div_res_transmission(xs, ys, exp_array, plot=False):
    """Loss function for fit qubit spectra by the Kullback-Leibler divergence.

    Args:
        data: dateset from experiment.
    """
    # Normalize
    exp_array = jnp.abs(10 * jnp.log10(exp_array))  # unit in db
    data_exp = jnp.nan_to_num(exp_array, nan=jnp.nanmin(exp_array))
    data_exp = jnp.abs(data_exp - jnp.nanmean(data_exp, axis=1, keepdims=True))
    data_exp = data_exp / jnp.nansum(data_exp)
    if plot:
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].pcolormesh(xs, ys, data_exp.T)
        axs[0].set_title(r'$\mathbf{P}(\varepsilon,b)$')

    # convert bias to phi_ext/phi_0
    phi_list, ys = wrapper_phi_freq(xs, ys)

    spec_out = _cal_transmission_vs_phi(phi_list, 1)
    peak_list = spec_out[:, 0]

    # set a bandwidth
    bandwidth = hk.get_parameter('bandwidth', [], init=jnp.zeros)
    noise = hk.get_parameter('noise', [], init=jnp.zeros)
    sim_data = []
    for peak in peak_list:
        sim_data.append(cauchy.pdf(ys, loc=peak, scale=bandwidth))
    sim_data = jnp.asarray(sim_data)

    # add noise
    sim_data += noise
    sim_data = sim_data / jnp.sum(sim_data)

    # plot
    if plot:
        axs[1].pcolormesh(xs, ys, sim_data.T)
        axs[1].set_title(r'$\mathbf{Q}(\varepsilon,\varphi)$')
        plt.show()
    # calculate kl div
    return jnp.nansum(data_exp * jnp.log(data_exp / sim_data))


@hk.without_apply_rng
@hk.transform
def loss_fit_transmission(bias_list,
                          freq_list,
                          qubit_level,
                          num_photon,
                          periodic=True):
    """Loss function for fit cavity-qubit transmission spectrum.

    Args:
        bias_list (list[~numpy.ndarray]): Bias data.
        freq_list (list[~numpy.ndarray]): Spectrum data.
        qubit_level (int): Qubit energy level for transmission spectrum.
        num_photon (int): The number of photon in cavity.
        periodic (bool): Using periodic boundary condition of phi
            when the target phi is beyond the phi basis range.
    """
    phi_list, freq_list = wrapper_phi_freq(bias_list, freq_list, periodic)

    spec_out = _cal_transmission_vs_phi(phi_list, num_photon)
    sim_freq_list = spec_out[:, qubit_level]
    # use mean square error with l2-norm
    # norm = jnp.linalg.norm(freq_list - jnp.mean(freq_list))

    return truncated_mse(sim_freq_list, freq_list) * 1e4



from typing import Callable, Optional, List, Union
from functools import reduce
import json
import numbers
import numpy as np
import jax
import jax.numpy as jnp
import scipy.constants as const
import pprint

import supergrad


def _promote_to_zero_list(arg, length):
    """
    Ensure `arg` is a list of length `length`.  If `arg` is None it is promoted
    to `[0]*length`.  All other inputs are checked that they match the correct
    form.

    Returns:
        list_ : list
            A list of integers of length `length`.
    """
    if arg is None:
        arg = [0] * length
    elif not isinstance(arg, list):
        arg = [arg]
    if not len(arg) == length:
        raise ValueError("All list inputs must be the same length.")
    return arg


def basis(dimensions, n=None, offset=None):
    """Generates the vector representation of a Fock state.

    Args:
        dimensions: int or list of ints
            Number of Fock states in Hilbert space.  If a list, then the resultant
            object will be a tensor product over spaces with those dimensions.

        n: int or list of ints, optional (default 0 for all dimensions)
            Integer corresponding to desired number state, defaults to 0 for all
            dimensions if omitted.  The shape must match ``dimensions``, e.g. if
            ``dimensions`` is a list, then ``n`` must either be omitted or a list
            of equal length.

        offset: int or list of ints, optional (default 0 for all dimensions)
            The lowest number state that is included in the finite number state
            representation of the state in the relevant dimension.

    Returns:
        state: `ndarray` representing the requested number state ``|n>``.
    """
    # Promote all parameters to lists to simplify later logic.
    if not isinstance(dimensions, list):
        dimensions = [dimensions]
    n_dimensions = len(dimensions)
    # promote None to zero list
    if n is None:
        n = [0] * n_dimensions
    if offset is None:
        offset = [0] * n_dimensions
    ns = [
        m - off for m, off in zip(_promote_to_zero_list(n, n_dimensions),
                                  _promote_to_zero_list(offset, n_dimensions))
    ]
    if any((not isinstance(x, numbers.Integral)) or x < 0 for x in dimensions):
        raise ValueError("All dimensions must be >= 0.")
    if not all(0 <= n < dimension for n, dimension in zip(ns, dimensions)):
        raise ValueError("All basis indices must be "
                         "`offset <= n < dimension+offset`.")
    # construct Fock basis in KronObj
    location, size = 0, 1
    for m, dimension in zip(reversed(ns), reversed(dimensions)):
        location += m * size
        size *= dimension
    psi = np.zeros(size, dtype=complex)
    psi[location] = 1
    return psi.reshape((-1, 1))


def qutrit_basis():
    """Basis states for a three level system (qutrit)

    Returns:
        qstates : array
            Array of qutrit basis vectors

    """
    out = np.empty((3,), dtype=object)
    out[:] = [basis(3, 0), basis(3, 1), basis(3, 2)]
    return out


def ghz_state(n=3):
    """
    Returns the N-qubit GHZ-state.

    Args:
        N : int (default=3)
            Number of qubits in state

    Returns:
        array: N-qubit GHZ-state
    """
    state = (tensor(*[basis(2) for k in range(n)]) +
             tensor(*[basis(2, 1) for k in range(n)]))
    return state / jnp.sqrt(2)


def fock_dm(dimensions, n=None, offset=None):
    """Density matrix representation of a Fock state

    Constructed via outer product of :func:`supergrad.utils.utility.basis`.

    Args:
        dimensions : int or list of ints
            Number of Fock states in Hilbert space.  If a list, then the resultant
            object will be a tensor product over spaces with those dimensions.

        n : int or list of ints, optional (default 0 for all dimensions)
            Integer corresponding to desired number state, defaults to 0 for all
            dimensions if omitted.  The shape must match ``dimensions``, e.g. if
            ``dimensions`` is a list, then ``n`` must either be omitted or a list
            of equal length.

        offset : int or list of ints, optional (default 0 for all dimensions)
            The lowest number state that is included in the finite number state
            representation of the state in the relevant dimension.

    Returns:
        dm : ndarray
            Density matrix representation of Fock state.
    """
    psi = basis(dimensions, n, offset=offset)

    return psi * np.conj(psi).T


def create_state(ar_truncated_dim: np.ndarray, ar_ix: np.ndarray):
    """Creates N-Q basis.

    Args:
        ar_truncated_dim: the number of bands in each tensor
        ar_ix: the index of band in each tensor

    Returns:
        Array as the state `|` i1 i2 i3...>
    """
    return tensor_np(*[basis(x, ix) for x, ix in zip(ar_truncated_dim, ar_ix)])


def create_state_init(ar_truncated_dim: np.ndarray,
                      ar_ix_max: Union[np.ndarray, List[List[int]]]):
    """Creates initial states in a N-Q system.

    Args:
        ar_truncated_dim: the number of bands in each tensor
        ar_ix_max: [N,2] array as (qubit index, band maximum band index)
            to iterate all bands from 0 to maximum.
            [[0, 2], [1,2]] will create `|` 00> `|` 01> `|` 10> `|` 11> states.

    Returns:
        a list of initial states
    """
    ar_truncated_dim = np.asarray(ar_truncated_dim)
    list_max = np.ones(ar_truncated_dim.size, dtype=int)
    for ix, x in ar_ix_max:
        list_max[ix] = x

    ar_ix = np.stack(np.meshgrid(*[np.arange(x) for x in list_max],
                                 indexing="ij"),
                     axis=-1).reshape((-1, ar_truncated_dim.size))

    list_state_init = [
        basis(list(ar_truncated_dim), list(band_ix)) for band_ix in ar_ix
    ]
    return np.array(list_state_init), ar_ix


def compute_average_photon(power: float,
                           freq: float,
                           qc: Optional[float] = None,
                           kappa_c: Optional[float] = None) -> float:
    """Computes the average photon number in a cavity.

    Note, must specify one and only one of qc and kappa_c.

    Args:
        power (float): the input power, in unit dBm
        freq (float): the cavity frequency (normal frequency), in unit GHz
        qc (float): the coupling quality factor
        kappa_c (float): the coupling decay rate (normal frequency)

    Returns:
        float: The average photon number.
    """

    if (qc is not None and kappa_c is not None) or (qc is None and
                                                    kappa_c is None):
        raise ValueError("Must specify one and only one of qc and kappa_c")

    # Convert to joule/s
    power = 10**(power / 10) * 1e-3
    # Convert to kappa
    if kappa_c is None:
        kappa_c = freq * 1e9 / qc

    return power / (freq * const.h * 1e9) / kappa_c / 2 / np.pi


def tensor(*operators):
    """Calculates the tensor product of input operators by `jax.numpy.kron`.
    """
    return reduce(jnp.kron, operators)


def tensor_np(*operators):
    """Calculates the tensor product of input operators by `numpy.kron`.
    """
    return reduce(np.kron, operators)


def _parse_operator(operator, **kwargs) -> jnp.ndarray:

    if isinstance(operator, Callable):
        return operator(**kwargs)
    if isinstance(operator, jnp.ndarray):
        return operator
    raise TypeError("Unsupported operator type: ", type(operator))


def identity_wrap(
    operator: Union[Callable, jnp.ndarray],
    subsystem_list,
    subsystem=None,
    coeff=1,
    **kwargs,
) -> jnp.ndarray:
    """Wrap given operator in subspace `subsystem` in identity operators to form
    full Hilbert-space operator.

    Args:
        operator:
            operator acting in Hilbert space of `subsystem`
        subsystem_list (list):
            list of all subsystems relevant to the Hilbert space.
        subsystem (string):
            subsystem where diagonal operator is defined. set `None` if operator
            is `Callable`
        coeff (float): coefficient of operator
        kwargs:
            kwargs for calling operator.
    """
    if subsystem is None:
        if isinstance(operator, Callable):
            subsystem = operator.__self__
        else:
            raise ValueError('Unsupported `subsystem` argument')
    operator = _parse_operator(operator, **kwargs)  # Call operator
    dims = [the_subsys.dim for the_subsys in subsystem_list]
    subsystem_index = subsystem_list.index(subsystem)

    return supergrad.KronObj([coeff * operator], dims, [subsystem_index])


def permute(unitary, dims, order):
    """ Permutes the tensor structure of a composite object in the given order.

    Args:
        unitary (array): 2d-array of a composite object
        dims (list): list of int
            dimensions of a composite object
        order (list): list of int
            the given order of the tensor structure. Note that `len(dims)` == `len(order)`,
            and the order in range [0, `len(dims)`-1]
    """
    # construct subscripts and operands
    unitary_dim = list(dims)
    shape = unitary.shape
    unitary_tensor = unitary.reshape(unitary_dim * 2)
    script = list(np.arange(len(order)) + 1) + list(-1 * np.arange(len(order)) -
                                                    1)
    locs = [loc + 1 for loc in order]  # index start from 1
    aux_locs = [-loc for loc in locs]
    target_script = locs + aux_locs
    return jnp.einsum(unitary_tensor, script, target_script).reshape(shape)


def reduced_unitary(unitary, dims, partial_trace):
    """Calculate the partial trace to extract the unitary of target subsystem,
    when dealing with composite systems.

    Args:
        unitary (array): 2d-array of a composite object
        dims (list): list of int
            dimensions of a composite object
        partial_trace (list): list of int
            the order of subsystem will be traced. Note the order in range
            [0, `len(dims)`-1]
    """
    unitary_dim = list(dims)
    assert np.prod(dims) == unitary.shape[0]
    unitary_tensor = unitary.reshape(unitary_dim * 2)
    script = list(np.arange(len(dims)) + 1)
    script_2 = list(-1 * np.arange(len(dims)) - 1)
    locs = list(np.arange(len(dims)))
    for idx in sorted(partial_trace, reverse=True):
        script_2[idx] *= -1  # set script for calculating trace
        locs.pop(idx)
        dims.pop(idx)
    target_script = [loc + 1 for loc in locs] + [-loc - 1 for loc in locs]
    return jnp.einsum(unitary_tensor, script + script_2, target_script).reshape(
        np.prod(dims), np.prod(dims)) / 2**len(partial_trace)


def const_init(val):
    """
    return an intializer which generates constant value or arrays

    Args:
        val: the constant value

    Returns:
        A callable of shape, dtype to generate a constant value or array.
    """

    if val is None:
        val = 1

    def initializer(shape, dtype):
        if shape:
            return jnp.full(shape, val, dtype)
        else:
            return jnp.array(val, dtype=dtype)

    return initializer


def convert_device_array(x):
    """Convert a device array to a list or a number.
    This is for convert parameters to json compatible format.
    """

    if x.size == 1:
        return float(x)
    else:
        return x.tolist()


def convert_to_json_compatible(x):
    """
    Convert input to json compatible format.
    In particular, it will replace DeviceArray or np.array in the input.
    """

    return jax.tree_util.tree_map(convert_device_array, x)


def convert_to_haiku_dict(x):
    """
    Convert input to haiku compatible dictionary.
    In particular, it will replace input float to the DeviceArray.
    """

    def _parse_params(x):
        if isinstance(x, float):
            return jnp.array(x)
        else:
            return x

    return jax.tree_util.tree_map(_parse_params, x)


def dump_params(params, fp):
    """Dump parameters to json."""
    json.dump(convert_to_json_compatible(params), fp)


def load_params(fp):
    """Load parameters from json."""
    return convert_to_haiku_dict(json.load(fp))


def tree_print(t):
    """Print jax pytree in a human readable way."""

    pp = pprint.PrettyPrinter(indent=2)
    pp.pprint(convert_to_json_compatible(t))



import jax.numpy as jnp
import networkx as nx

from supergrad.scgraph.graph import SCGraph

# Before we multiply 2pi, ec, ej are in units of Ghz
# We multiply 2pi to them so that we follow the same convention of setting
# hbar=1 as done in qutip's sesolve
transmon_1 = {
    'ec': jnp.array(0.105 * 2 * jnp.pi),
    'ej': jnp.array(33.063 * 2 * jnp.pi),
    'system_type': 'transmon',
    'shared_param_mark': 1,
    'arguments': {
        'ng': 0.5,
        'num_basis': 62,
        'n_max': 31,
        'phiext': 0.,
    }
}
transmon_2 = {
    'ec': jnp.array(0.120 * 2 * jnp.pi),
    'ej': jnp.array(22.234 * 2 * jnp.pi),
    'system_type': 'transmon',
    'shared_param_mark': 2,
    'arguments': {
        'ng': 0.5,
        'num_basis': 62,
        'n_max': 31,
        'phiext': 0.,
    }
}
coupler_1 = {
    'ec': jnp.array(0.185 * 2 * jnp.pi),
    'ej': jnp.array(30.694 * 2 * jnp.pi),
    'system_type': 'transmon',
    'shared_param_mark': 3,
    'arguments': {
        'ng': 0.5,
        'num_basis': 62,
        'n_max': 31,
        'truncated_dim': 3,
        'd': 0,
        'phiext': 0.,
    }
}
g_12 = {
    'capacitive_coupling': {
        'strength': jnp.array(12e-3 * 2 * jnp.pi)
    },
}
g_1c = {
    'capacitive_coupling': {
        'strength': jnp.array(122e-3 * 2 * jnp.pi)
    },
}
g_2c = {
    'capacitive_coupling': {
        'strength': jnp.array(105e-3 * 2 * jnp.pi)
    },
}


class Tmon1D(SCGraph):
    """Device parameters for 4 qubits-4 couplers loop quantum processor graph.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(None)
        # initialize graph
        temp_graph = nx.Graph()
        # adding attributes to nodes
        params = [transmon_1, coupler_1, transmon_2, coupler_1]
        for i in range(8):
            temp_graph.add_node(i, **params[i % 4])

        # adding attributes to edges
        tc_couplings = {
            0: [g_1c, g_12],
            1: [g_2c, None],
            2: [g_2c, g_12],
            3: [g_1c, None]
        }
        for i in range(8):
            for j, coupling in enumerate(tc_couplings[i % 4], 1):
                if coupling is not None:
                    temp_graph.add_edge(i, (i + j) % 8, **coupling)

        # relabel nodes
        label_mapping = dict(
            (label, ''.join(['q', str(label)])) for label in temp_graph.nodes)
        temp_graph = nx.relabel_nodes(temp_graph, label_mapping)
        # save temp_graph
        self.add_nodes_from(temp_graph.nodes.data())
        self.add_edges_from(temp_graph.edges.data())

        if seed is not None:
            self.add_lcj_params_variance_to_graph(seed=seed)


class XGateTmon1D(Tmon1D):
    """Add pulse parameters for simultaneous X gates.
    In the example, we add 3 pulses on `q0`, `q2` and `q4`.

    Args:
        seed : (int) The seed for random device parameters variance.
            default seed is None, means no variance.
    """

    def __init__(self, seed=None):
        super().__init__(seed)
        self.add_node(
            'q0', **{
                'pulse': {
                    'amp': jnp.array(0.0922632),
                    'length': jnp.array(39.99841052),
                    'omega_d': jnp.array(31.89213402),
                    'phase': jnp.array(-0.06459036),
                    'pulse_type': 'cos',
                    'operator_type': 'n_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.0086247),
                    'post_comp': jnp.array(-0.0086247)
                }
            })
        self.add_node(
            'q2', **{
                'pulse': {
                    'amp': jnp.array(0.10390872),
                    'length': jnp.array(39.92211365),
                    'omega_d': jnp.array(27.99554391),
                    'phase': jnp.array(0.05805683),
                    'pulse_type': 'cos',
                    'operator_type': 'n_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.3714724),
                    'post_comp': jnp.array(-0.3714724)
                }
            })
        self.add_node(
            'q4', **{
                'pulse': {
                    'amp': jnp.array(0.09196213),
                    'length': jnp.array(39.88357291),
                    'omega_d': jnp.array(31.97277349),
                    'phase': jnp.array(-0.07858071),
                    'pulse_type': 'cos',
                    'operator_type': 'n_operator',
                    'delay': 0.
                },
                'compensation': {
                    'pre_comp': jnp.array(0.79958712),
                    'post_comp': jnp.array(-0.79958712)
                }
            })



import copy
import jax
import jax.numpy as jnp
from jax.experimental.ode import odeint

from supergrad.quantum_system import KronObj, LindbladObj
from supergrad.time_evolution.ode import _parse_hamiltonian, ode_expm
from supergrad.utils.sharding import get_sharding


def mesolve(hamiltonian,
            rho0,
            tlist,
            c_ops=None,
            args=None,
            options={},
            solver='ode_expm'):
    """
    Master equation evolution of a density matrix for a given Hamiltonian and
    set of collapse operators, or a Liouvillian.

    Evolve the density matrix (`rho0`) using a given Hamiltonian or
    Liouvillian (`LindbladObj`) and an optional set of collapse operators
    (`c_ops`), by integrating the set of ordinary differential equations
    that define the system. In the absence of collapse operators the system is
    evolved according to the unitary evolution of the Hamiltonian.

    The output is either the density matrix at arbitrary points in time
    (`tlist`).

    Args:
        hamiltonian (list, array):
            System Hamiltonian as a :obj:`KronObj` , list of :obj:`KronObj` and
            coefficient. List format and options can be found in description.
            or alternatively a system Liouvillian :obj:`LindbladObj`.

        rho0 (array):
            Initial density matrix (rho) or initial set of density matrices.
            Alternatively evolve a set of density matrix use vmap transformation.

        tlist (list, array):
            List of times for :math:`t`.

        c_ops : None / list of :class:`KronObj` / list of :class:`LindbladObj`
            single collapse operator, or list of collapse operators, or a list
            of Liouvillian super-operators.

        args (dict, optional):
            Dictionary of scope parameters for time-dependent Hamiltonian.

        options (dict, optional):
            `multi_device` (bool):
            `True` for enable multi-device parallelism evolution support,
            if many states are evolved simultaneously, the compiler can
            automatically batch the states to different devices.
            `diag_ops` (bool):
            It is only for `ode_expm`, if `True` the Hermitian super-operator
            will be convert to diagonal form before exponentiation.

        solver (str, optional): select differentiable ODE solver.
            `odeint` for using Dormand-Prince ODE integration,
            `ode_expm` for using the matrix exponential at each time step.

    Returns: Values of the solution density matrix at each time point in `tlist`,
        represented as an array with the same shape/structure as `rho0` except
        with a new leading axis of length `len(t)`. if a set of states be evolved
        simultaneously, the returns with another new leading axis of length `len(rho0)`.
    """
    diag_ops = options.pop('diag_ops', False)
    if diag_ops:
        assert solver == 'ode_expm'

    # convert options to hashable tuple
    options = tuple(options.items())

    if isinstance(hamiltonian, (KronObj, jnp.ndarray, list)):
        h_td = _parse_hamiltonian(hamiltonian, tlist, args, diag_ops)
    else:
        raise ValueError('Invalid Hamiltonian type.')
    # Construct Liouvillian
    lind = LindbladObj()
    if c_ops is not None:
        for c_op in c_ops:
            if isinstance(c_op, LindbladObj):
                lind += c_op
            if isinstance(c_op, list):
                lind.add_lindblad_operator(*c_op)
            else:
                lind.add_lindblad_operator(c_op)
        if diag_ops:
            lind._diagonalize_super_operator()
    if rho0.ndim == 2:
        return _mesolve(rho0, tlist, h_td, lind, args, solver, options)
    elif rho0.ndim == 3:
        _simd_mesolve = jax.vmap(_mesolve,
                                 in_axes=(0, None, None, None, None, None,
                                          None))
        rho0 = jax.lax.with_sharding_constraint(rho0,
                                                get_sharding('p', None, None))
        return _simd_mesolve(rho0, tlist, h_td, lind, args, solver, options)
    else:
        raise ValueError(
            '`rho0` is array or array of arrays representing the initial density'
        )


def _mesolve(rho0, tlist, h_td, lind: LindbladObj, args, solver, options):
    """The private function of `mesolve`"""

    tlist = jnp.asarray(tlist)
    # Start evolution
    if solver == 'ode_expm':

        def _func(rho, t):
            ham = h_td(t, args=args)
            lind_td = copy.deepcopy(lind)
            lind_td.add_liouvillian(ham)
            return lind_td

        rho_evo = ode_expm(_func, rho0, tlist, **options)
    elif solver == 'odeint':
        if isinstance(rho0, KronObj):
            rho0 = rho0.full()

        def _func(rho, t):
            ham = h_td(t, args=args)
            lind_td = copy.deepcopy(lind)
            lind_td.add_liouvillian(ham)
            return lind_td @ rho

        rho_evo = odeint(_func, rho0, tlist, **options)
    else:
        raise ValueError('Unknown ode solver.')

    return rho_evo


def mesolve_final_states_w_basis_trans(hamiltonian,
                                       rho_list,
                                       tlist,
                                       transform_matrix=None,
                                       **kwargs):
    """Compute the final density matrices and applying basis transformation
    according to the transform_matrix.

    Args:
        hamiltonian: `jax.numpy.ndarray`, list
            System Hamiltonian as a :obj:`~ndarray` , list of :obj:`ndarray` and
            coefficient. List format is the same as `qutip.mesolve`.

        rho_list: `jax.numpy.ndarray`
            Initial density matrix (rho) or array of density matrices.

        tlist: array_like of float
            List of times for :math:`t`.

        transform_matrix: `jax.numpy.ndarray`
            The transform unitary from product basis to the desired basis.

        kwargs:
            Keyword arguments will be pass to `supergrad.time_evolution.mesolve`
    """
    # evolution in the product basis with cycle transform
    if transform_matrix is not None:
        # Evolve in multi-qubit eigenbasis

        @jax.jit
        def _evolving_rhos(transform_matrix, rho_list):
            # Fix bug in gradient computation with multi-device
            # we use jit to tell the compiler transform_matrix must be provided
            # before the evolution
            rho_list = transform_matrix @ rho_list @ transform_matrix.conjugate(
            ).transpose()
            res = mesolve(hamiltonian, rho_list, tlist, **kwargs)
            return jnp.conj(transform_matrix).T @ res[:,
                                                      -1, :] @ transform_matrix

        density_array = _evolving_rhos(transform_matrix, rho_list)
    else:
        # Evolve in multi-qubit product basis
        res = mesolve(hamiltonian, rho_list, tlist, **kwargs)
        density_array = res[:, -1, :]
    return density_array



from abc import ABC, abstractmethod
import jax.scipy as jsp
import haiku as hk


class QuantumSystem(ABC, hk.Module):
    """Generic quantum system class.
    """

    def __init__(self, name: str = 'quantum_system'):
        super().__init__(name=name)

        self._sys_type = type(self).__name__

    @property
    @abstractmethod
    def dim(self) -> int:
        """Return dimension of Hilbert space"""

    @abstractmethod
    def idling_hamiltonian(self):
        """Time independent part of Hamiltonian for QuantumSystem.
        """

    @abstractmethod
    def eigenenergies(self):
        """Eigenenergies of Quantum system.
        """

    def _calc_evals(self):
        """Calculate quantum system's eigenenergy using `jax.scipy.linalg.eigh`.

        """

        h = self.idling_hamiltonian()
        eigval = jsp.linalg.eigh(h, eigvals_only=True)
        return eigval

    def _calc_eigsys(self):
        """Calculate quantum system's eigenenergy and eigenvectors using
        `jax.scipy.linalg.eigh`.
        """

        h = self.idling_hamiltonian()
        eigval, eigvec = jsp.linalg.eigh(h, eigvals_only=False)
        return eigval, eigvec



from supergrad.helper.helper import Helper
from supergrad.helper.compute_unitary_evolve import Evolve
from supergrad.helper.compute_spectrum import Spectrum

__all__ = ['Helper', 'Evolve', 'Spectrum']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Author: `Ziang Wang`
Date: 2022-07-12 09:23:17
LastEditTime: 2022-07-12 09:23:17
"""
import importlib
import os
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
project_name = 'supergrad'


def load_version_module(pkg_path):
    spec = importlib.util.spec_from_file_location(
        'version', os.path.join(pkg_path, 'version.py'))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


_version_module = load_version_module(project_name)
__version__ = _version_module._get_version_for_build()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='supergrad',
    version=__version__,
    author='IQUBIT',
    url='https://github.com/iqubit-org/supergrad',
    description='SuperGrad differentiable Hamiltonian simulator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["examples", "tests"]),
    package_data={},
    python_requires='>=3.9',
    platforms='any',
    install_requires=[
        'numpy', 'scipy', 'matplotlib', 'jax', 'dm_haiku', 'jaxopt', 'networkx',
        'tqdm', 'rich'
    ],
    extras_require={
        # Used for benchmark
        'benchmark': [
            'qiskit_dynamics', 'qutip', 'pytest', 'pytest-benchmark',
            'scqubits', 'psutil'
        ],
        'docs': ['qutip', 'qutip-qip'],
        # Used for generate paper
        'latex': ['pylatex'],
    },
    license='Apache-2.0',
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
)



from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt
import jax
import jax.numpy as jnp
from jax.scipy.special import erf
import haiku as hk


class PulseBase(hk.Module, ABC):
    """Generic pulse for various gate operations.

    Args:
        length (float): the length of pulse, set `None` for automatically collect
            parameter by haiku.
        amp (float): the max amplitude of pulse, set `None` for automatically
            collect parameter by haiku.
        delay (float): the delay time of pulse
        modulate_wave (bool): modulate the wave packet or not
        name (string): the name of module
    """

    def __init__(self,
                 length=None,
                 amp=None,
                 delay=0.,
                 modulate_wave=False,
                 name: str = 'pulse',
                 **kwargs):
        super().__init__(name=name)

        if length is None:
            self.length = hk.get_parameter('length', [], init=jnp.ones)
        else:
            self.length = length
        if amp is None:
            self.amp = hk.get_parameter('amp', [], init=jnp.ones)
        else:
            self.amp = amp
        self.delay = delay
        self.modulate_wave = modulate_wave
        if self.modulate_wave:
            self.omega_d = hk.get_parameter('omega_d', [], init=jnp.ones)
            self.phase = hk.get_parameter('phase', [], init=jnp.zeros)

    @abstractmethod
    def create_envelope_pulse(self, t, args={}):
        """Create envelope pulse shape for various gate operations."""

    def create_pulse(self, t, args={}):
        """Create desired pulse shape for various gate operations."""
        t_pulse = t - self.delay
        shape = self.create_envelope_pulse(t, args)
        if self.modulate_wave:
            shape *= jnp.cos(self.omega_d * t_pulse + self.phase)
        return shape

    @property
    def pulse_endtime(self):
        """The time of last pulse frame"""
        return self.length + self.delay


class PulseTrapezoid(PulseBase):
    """The trapezoid pulse.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        amp: max amplitude of the trapezoid
        t_ramp: time of the rising and falling edges
        t_plateau: the holding time of the plateau of the waveform
        length: the length of pulse shape
    """

    def __init__(
            self,
            length=False,  # unused argument
            amp=None,
            delay=0,
            modulate_wave=False,
            name: str = 'pulse_trapezoid',
            **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        # construct activation function
        self.t_ramp = hk.get_parameter(
            't_ramp', [], init=jnp.ones)  # time for rising and falling edge
        self.t_plateau = hk.get_parameter('t_plateau', [], init=jnp.ones)
        self.length = self.t_plateau + 2 * self.t_ramp

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        time1 = jax.nn.relu(t_pulse / self.t_ramp)
        time2 = jax.nn.relu(
            (2 * self.t_ramp + self.t_plateau - t_pulse) / self.t_ramp)
        shape = self.amp * jnp.min(
            jnp.array([jnp.ones_like(time1), time1, time2]), axis=0)

        return shape * ((t_pulse >= 0) & (t_pulse <= 0 + self.length))


class PulseCosineRamping(PulseBase):
    """A flat-top pulse with cosine ramping at both ends.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        amp: max amplitude of the flat-top
        t_ramp: time of the rising and falling edges
        t_plateau: the holding time of the plateau of the waveform
        length: the length of pulse shape
    """

    def __init__(
            self,
            length=False,  # unused argument
            amp=None,
            delay=0,
            modulate_wave=False,
            name: str = 'pulse_rampcos',
            **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        # construct activation function
        self.t_ramp = hk.get_parameter(
            't_ramp', [], init=jnp.ones)  # time for rising and falling edge
        self.t_plateau = hk.get_parameter('t_plateau', [], init=jnp.ones)
        self.length = self.t_plateau + 2 * self.t_ramp

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        pulse_1 = (1 - jnp.cos(jnp.pi * t_pulse / self.t_ramp)) / 2 * (
            t_pulse < self.t_ramp)
        pulse_2 = (t_pulse >= self.t_ramp) & (t_pulse
                                              <= self.t_ramp + self.t_plateau)
        pulse_3 = (1 - jnp.cos(jnp.pi *
                               (self.length - t_pulse) / self.t_ramp)) / 2 * (
                                   t_pulse > self.t_ramp + self.t_plateau)
        shape = self.amp * (pulse_1 + pulse_2 + pulse_3)

        return shape * ((t_pulse >= 0) & (t_pulse <= 0 + self.length))


class PulseGaussian(PulseBase):
    """Creates normalized Gaussian pulse, center at center of time range.
    The duration of the pulse is determined by ``2*cutoff*sigma``.
    If ``length < 2*cutoff*sigma``, the waveform will be truncated.
    If ``length > 2*cutoff*sigma``, zeros will be appended to the waveform.

    Note head/tail are directly truncated

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        sigma: the sigma of Gaussian function
        cutoff: the cutoff of Gaussian function

    Returns:
        an array of amplitudes
    """

    def __init__(
            self,
            length=False,  # unused argument
            amp=None,
            delay=0,
            modulate_wave=False,
            name: str = 'pulse_gaussian',
            **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        self.sigma = hk.get_parameter('sigma', [], init=jnp.ones)
        self.cutoff = hk.get_parameter('cutoff', [], init=jnp.ones)
        self.length = 2 * self.cutoff * self.sigma

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        t1 = self.cutoff * self.sigma
        offset = jnp.exp(-self.cutoff**2 / 2)

        shape = (jnp.exp(-(t_pulse - t1)**2 / self.sigma**2 / 2) -
                 offset) / (1 - offset)

        shape = shape * ((t_pulse >= 0) & (t_pulse <= self.length)) * self.amp

        return shape


class PulseGaussianSquare(PulseBase):
    """Creates Gaussian square pulse shape.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations, unit in ns
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        sigma: the sigma of Gaussian square
        cutoff: the cutoff of Gaussian square

    Returns:
        an array of amplitudes
    """

    def __init__(self,
                 length=None,
                 amp=None,
                 delay=0,
                 modulate_wave=False,
                 name: str = 'pulse_gaussian_envelope',
                 **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        self.sigma = hk.get_parameter('sigma', [], init=jnp.ones)
        self.cutoff = hk.get_parameter('cutoff', [], init=jnp.ones)

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        t1 = self.cutoff * self.sigma
        t2 = self.length - self.cutoff * self.sigma
        offset = jnp.exp(-self.cutoff**2 / 2)

        shape = (jnp.exp(-(t_pulse - t1)**2 / self.sigma**2 / 2) -
                 offset) / (1 - offset) * (t_pulse <= t1)
        shape += 1.0 * ((t_pulse > t1) & (t_pulse <= t2))
        shape += (jnp.exp(-(t_pulse - t2)**2 / self.sigma**2 / 2) -
                  offset) / (1 - offset) * (t_pulse > t2)

        shape = shape * ((t_pulse >= 0) & (t_pulse <= self.length)) * self.amp

        return shape


class PulseCosine(PulseBase):
    """Creates cosine style pulse from phase 0 to 2pi,

    This is centered at the center of time range normalized.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations, unit in ns
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Returns:
        an array of amplitudes
    """

    def __init__(self,
                 length=None,
                 amp=None,
                 delay=0,
                 modulate_wave=False,
                 name: str = 'pulse_1mcos',
                 **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        i_quad = 0.5 * (1 + jnp.cos(2 * jnp.pi / self.length *
                                    (t_pulse - self.length / 2)))

        shape = (i_quad) * (t_pulse >= 0) * (t_pulse <= self.length) * self.amp

        return shape


class PulseTanh(PulseBase):
    """Creates tanh up/down pulse.

    Note:
    - The pulse shape will reach a flat top of value 1 only when
    ``length`` is large enough, e.g., ``length > (10 + 2*cutoff) * sigma``.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations, unit in ns
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        sigma: the sigma of tanh function
        cutoff: the cutoff of tanh function

    Returns:
        an array of amplitudes
    """

    def __init__(self,
                 length=None,
                 amp=None,
                 delay=0,
                 modulate_wave=False,
                 name: str = 'pulse_tanh',
                 **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        self.sigma = hk.get_parameter('sigma', [], init=jnp.ones)
        self.cutoff = hk.get_parameter('cutoff', [], init=jnp.ones)

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        t1 = self.cutoff * self.sigma
        t2 = self.length - self.cutoff * self.sigma
        offset = 0.5 * (jnp.tanh(-1.0 * self.cutoff) +
                        jnp.tanh(self.length / self.sigma - self.cutoff))
        peak = jnp.tanh(self.length / 2 / self.sigma - self.cutoff)

        shape = (0.5 * jnp.tanh((t_pulse - t1) / self.sigma) + 0.5 * jnp.tanh(
            (t2 - t_pulse) / self.sigma) - offset) / (peak - offset)

        shape = shape * ((t_pulse >= 0) & (t_pulse <= self.length)) * self.amp

        return shape


class PulseErf(PulseBase):
    """Creates error function up/down pulse, integral 1.

    Both up and down parts are erf style, controlled by sigma.
    the erf function is truncated at sigma=-cutoff(as t=0)
    and sigma=+cutoff(as t=tmax)
    The pulse is nearly flat in the middle if sigma is small.

    Args:
        length: the length of pulse
        amp: the amplitude of pulse
        delay: time delay for waiting other gate operations, unit in ns
        modulate_wave(bool): True for generate modulated sine wave
        name: module name

    Attributes:
        sigma: the sigma of erf function
        cutoff: the cutoff of erf function

    Returns:
        an array of amplitudes
    """

    def __init__(self,
                 length=None,
                 amp=None,
                 delay=0,
                 modulate_wave=False,
                 name: str = 'pulse_erf',
                 **kwargs):
        super().__init__(length, amp, delay, modulate_wave, name)

        self.sigma = hk.get_parameter('sigma', [], init=jnp.ones)
        self.cutoff = hk.get_parameter('cutoff', [], init=jnp.ones)

    def create_envelope_pulse(self, t, args={}):
        t_pulse = t - self.delay
        # t/2/sigma means the time corresponds to dx=1 in erf(x)
        sigma = self.length / 2 / self.sigma
        t1 = t_pulse - sigma * self.cutoff
        t2 = (self.length - t_pulse) - sigma * self.cutoff
        c1 = -self.cutoff
        c2 = -self.cutoff + self.length / sigma

        y = erf(t1 / sigma) + erf(t2 / sigma)
        y = y - erf(c1) - erf(c2)

        shape = y * self.amp / 2

        return shape * ((t_pulse >= 0) & (t_pulse <= self.length))


class PulseWithDRAG(PulseBase):
    """Creates pulse for single qubit operation.

    Args:
        envelope: pulse envelope
        delay: time delay for waiting other gate operations
        drive_dir: 0 for X and pi/2 for Y
        omega_d: driving frequency
        drag: using the derivative removal by adiabatic gate
        name: module name

    Attributes:
        length: the length of pulse shape
        amp: the max amplitude
        omega_d: the drive frequency
    """

    def __init__(self,
                 envelope: PulseBase = PulseCosine,
                 length=None,
                 amp=None,
                 delay=0,
                 modulate_wave=True,
                 name: str = 'pulse_drag',
                 **kwargs):
        assert modulate_wave is True
        super().__init__(length, amp, delay, modulate_wave, name)

        # DRAG codes
        self.lam = hk.get_parameter('lambda', [], init=jnp.zeros)

        self.envelope = envelope(length=self.length,
                                 amp=1.0,
                                 delay=delay,
                                 modulate_wave=False)
        # the length of pulse determinate by envelope
        self.length = self.envelope.length

    def create_envelope_pulse(self, t, args={}):
        envelope_pulse = self.amp * self.envelope.create_pulse(t)

        return envelope_pulse

    def create_pulse(self, t, args={}):
        """Overload `create_pulse` function."""
        t_pulse = t - self.delay
        shape = self.create_envelope_pulse(t, args)
        pulse = jnp.cos(self.omega_d * t_pulse + self.phase)
        output = shape * pulse

        if isinstance(t_pulse, (np.ndarray, jnp.ndarray)) and t_pulse.ndim > 0:
            d_envelope = jax.vmap(jax.grad(self.envelope.create_pulse))
        else:
            d_envelope = jax.grad(self.envelope.create_pulse)
        output -= self.lam * d_envelope(t) * jnp.sin(self.omega_d * t_pulse +
                                                     self.phase)
        return output * ((t_pulse >= 0) & (t_pulse <= 0 + self.length))


def draw(
    pulse: PulseBase,
    ts,
    function: Optional[str] = "pulse",
    title: Optional[str] = None,
):
    """Plot the pulse over an array `ts`.
    The ``function`` arg specifies which function to plot:

        - ``function == 'pulse'`` plots the full driving pulse.
        - ``function == 'envelope'`` plots the envelope.

    Args:
        ts: The time values array to plot.
        function: Which function to plot.
        title: Title of plot.
    """

    @hk.without_apply_rng
    @hk.transform
    def _draw(ts, function, title):
        if function == "pulse":
            y_vals = pulse.create_pulse(ts)
            title = title or "Value of " + pulse.name
        elif function == "envelope":
            y_vals = pulse.create_envelope_pulse(ts)
            title = title or "Envelope of " + pulse.name

        plt.figure(figsize=(8, 4))
        plt.plot(ts, y_vals)
        plt.title(title)
        plt.tight_layout()

    _draw.apply({}, ts, function, title)


if __name__ == '__main__':

    @hk.without_apply_rng
    @hk.transform
    def plot_pulse():
        pulse = PulseWithDRAG(delay=0, modulate_wave=True, name='pulse')
        ts = np.linspace(0, pulse.pulse_endtime + 4, 500)
        draw(pulse, ts, 'pulse')
        draw(pulse, ts, 'envelope')
        plt.show()

    plot_pulse.apply({
        'pulse': {
            'amp': jnp.array(0.5),
            'omega_d': jnp.array(2.),
            'phase': jnp.array(0.),
            't_ramp': jnp.array(30.),
            't_plateau': jnp.array(30.),
            'sigma': jnp.array(1.),
            'cutoff': jnp.array(2.),
            'length': jnp.array(60.),
            'lambda': jnp.array(0.)
        }
    })



from supergrad.quantum_system.base import QuantumSystem
from supergrad.quantum_system.artificial_atom import (CircuitLCJ, Transmon,
                                                      Fluxonium, Resonator)
from supergrad.quantum_system.interaction import (parse_interaction,
                                                  InteractingSystem,
                                                  InteractionTerm)
from supergrad.quantum_system.kronobj import KronObj
from supergrad.quantum_system.lindbladobj import LindbladObj

__all__ = [
    'QuantumSystem', 'Transmon', 'Fluxonium', 'Resonator', 'InteractingSystem',
    'KronObj', 'parse_interaction', 'LindbladObj', 'InteractionTerm',
    'CircuitLCJ'
]



# %%
import os
import numpy as np

import supergrad
from supergrad.quantum_system import Fluxonium, Transmon, InteractingSystem, Resonator, parse_interaction

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)


# %%
def test_composite_transmon():
    # load cached data
    spec_0 = np.load('data/data_composite_transmon_spec.npy')
    map_0 = np.load('data/data_composite_transmon_map.npy')

    class ExploreTransmon(supergrad.Helper):

        def _init_quantum_system(self):
            tm1 = Transmon(ej=40.,
                           ec=0.2,
                           d=0.1,
                           phiext=0.23 * 2 * np.pi,
                           ng=0.3,
                           n_max=40,
                           truncated_dim=3,
                           constant=True)
            tm2 = Transmon(ej=15.,
                           ec=0.15,
                           d=0.2,
                           phiext=0.,
                           ng=0.,
                           n_max=30,
                           truncated_dim=3,
                           constant=True)
            res = Resonator(f_res=4.5,
                            truncated_dim=4,
                            constant=True,
                            remove_zpe=True)

            coup1 = parse_interaction(g_strength=0.1,
                                      op1=tm1.n_operator,
                                      op2=res.creation_operator,
                                      add_hc=True,
                                      constant=True)
            coup2 = parse_interaction(g_strength=0.2,
                                      op1=tm2.n_operator,
                                      op2=res.creation_operator,
                                      add_hc=True,
                                      constant=True)
            self.hilbertspace = InteractingSystem([tm1, tm2, res],
                                                  [coup1, coup2])

        def energy_spectrum(self):
            return self.hilbertspace.eigenenergies(10)

        def energy_in_bare_indices(self):
            return self.hilbertspace.compute_energy_map()

    composite_tmon = ExploreTransmon()
    spec_1 = composite_tmon.energy_spectrum({})
    map_1 = composite_tmon.energy_in_bare_indices({})
    assert np.allclose(spec_0, spec_1)
    # remove NaN in the energy map
    idx = np.argwhere(~np.isnan(map_0))
    assert np.allclose(map_0[idx], map_1.flatten()[idx])


def test_composite_fluxonium():
    # load cached data
    spec_0 = np.load('data/data_composite_fluxonium_spec.npy')
    map_0 = np.load('data/data_composite_fluxonium_map.npy')

    class ExploreTransmon(supergrad.Helper):

        def _init_quantum_system(self):
            fm1 = Fluxonium(ej=3.5,
                            ec=1.6,
                            el=0.5,
                            phiext=0.33 * 2 * np.pi,
                            phi_max=5 * np.pi,
                            truncated_dim=4,
                            constant=True)
            coupler1 = Fluxonium(ej=4.0,
                                 ec=2.0,
                                 el=0.5,
                                 phiext=0.0,
                                 phi_max=5 * np.pi,
                                 truncated_dim=6,
                                 constant=True)
            fm2 = Fluxonium(ej=3.5,
                            ec=1.2,
                            el=0.5,
                            phiext=0.5 * 2 * np.pi,
                            phi_max=5 * np.pi,
                            truncated_dim=4,
                            constant=True)

            coup1 = parse_interaction(g_strength=0.04,
                                      op1=fm1.n_operator,
                                      op2=coupler1.n_operator,
                                      constant=True)
            coup2 = parse_interaction(g_strength=-0.02,
                                      op1=fm1.phi_operator,
                                      op2=coupler1.phi_operator,
                                      constant=True)
            coup3 = parse_interaction(g_strength=0.04,
                                      op1=fm2.n_operator,
                                      op2=coupler1.n_operator,
                                      constant=True)
            coup4 = parse_interaction(g_strength=-0.02,
                                      op1=fm2.phi_operator,
                                      op2=coupler1.phi_operator,
                                      constant=True)
            self.hilbertspace = InteractingSystem([fm1, coupler1, fm2],
                                                  [coup1, coup2, coup3, coup4])

        def energy_spectrum(self):
            return self.hilbertspace.eigenenergies(10)

        def energy_in_bare_indices(self):
            return self.hilbertspace.compute_energy_map()

    composite_tmon = ExploreTransmon()
    spec_1 = composite_tmon.energy_spectrum({})
    map_1 = composite_tmon.energy_in_bare_indices({})
    assert np.allclose(spec_0, spec_1)
    # remove NaN in the energy map
    idx = np.argwhere(~np.isnan(map_0))
    assert np.allclose(map_0[idx], map_1.flatten()[idx])



import argparse
import psutil
import time
import os
import sys
import signal
import subprocess


def max_memory_usage(pid):
    # Create a psutil process object
    ps_process = psutil.Process(pid)

    # Initialize max memory usage as 0
    max_mem = 0

    def handle_sigint(sig, frame):
        # Print the max memory usage
        print(max_mem)
        sys.exit(0)

    # Set the SIGINT handler
    signal.signal(signal.SIGINT, handle_sigint)

    # While the process is running
    while ps_process.is_running():
        try:
            # Get the memory info
            mem_info = ps_process.memory_info()

            # Update max memory usage
            max_mem = max(max_mem, mem_info.rss)

            # Sleep for a while
            time.sleep(0.01)

        except psutil.NoSuchProcess:
            break

    # Print the max memory usage
    print(max_mem)


def trace_max_memory_usage(func, pid):

    def wrapper(*args, **kwargs):
        process = subprocess.Popen(
            ['python', __file__, str(pid)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

        output = func(*args, **kwargs)
        # Stop the memory profiling process
        os.kill(process.pid, signal.SIGINT)
        # Get the stdout and stderr
        stdout_data, stderr_data = process.communicate()
        # delete \n in the end of the stdout
        stdout_data = stdout_data.decode('utf-8').strip()
        return output, int(stdout_data)

    return wrapper


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Memory profiling.')
    parser.add_argument('pid',
                        type=int,
                        help='the PID of the process to monitor')

    args = parser.parse_args()

    # Call the function with the PID from the command line
    max_memory_usage(args.pid)



"""The Kronecker Object (KronObj) class is designed to represent k-body Hamiltonian
in Kronecker product format. It's particularly useful for representing the sparse
total system Hamiltonian in a form convenient for computing Einstein summation.
"""
import numbers
from math import sqrt

import numpy as np
import jax
from jax.tree_util import register_pytree_node_class
import jax.numpy as jnp
import jax.scipy.linalg as jaxLA
import opt_einsum as oe

from supergrad.utils.utility import tensor, permute


@register_pytree_node_class
class KronObj(object):
    """The Kronecker Object (KronObj) class is designed to represent k-body Hamiltonian
    in Kronecker product format. It's particularly useful for representing the sparse
    total system Hamiltonian in a form convenient for computing Einstein summation.

    Additionally, the KronObj class supports mathematical operations such as addition(+)
    and multiplication(@) between KronObj instances.

    Args:
        inpt (list): list of array
            The local Hamiltonian that represents the Hamiltonian of each subsystem
            in a composite system. Each element of the `inpt` corresponds to
            the local Hamiltonian of the corresponding subsystem tagged in the `locs`.
        dims (list): list of int
            Dimensions of subsystems in the composited Hilbert Space.
        locs (list): list of int
            Location information of corresponding local Hamiltonian.
            For example, the first element of locs is the index of the first
            subsystem in `self.dims`, and so on.
        diag_unitary (array):
            Unitary of diagonal-unitary format.
        _nested_inpt (bool):
            For internal using only.
    """

    def __init__(self,
                 inpt=None,
                 dims=None,
                 locs=None,
                 _nested_inpt=False,
                 diag_unitary=None) -> None:

        if isinstance(inpt, (list, tuple, np.ndarray, jnp.ndarray)):
            # ndarray shape checker
            if _nested_inpt:
                # identify diagonal-unitary representation
                if diag_unitary is None:
                    self.diag_unitary = [None] * len(inpt)
                else:
                    self.diag_unitary = list(diag_unitary)
                # identify inpt location
                if locs is None:
                    self.locs = [None] * len(inpt)
                else:
                    self.locs = list(locs)
                if isinstance(inpt, (np.ndarray, jnp.ndarray)):
                    if inpt.ndim < 3:
                        raise ValueError('Unsupported data shape.')
                self._data = []
                if dims is None:
                    raise ValueError('No dimensions')
                else:
                    self.dims = dims
                for nest_list in inpt:
                    self._data.append([array for array in nest_list])
            else:
                # identify diag_unitary method
                if diag_unitary is None:
                    self.diag_unitary = [None]
                else:
                    self.diag_unitary = [diag_unitary]  # in the list form
                # identify localized method
                if locs is None:
                    self.locs = [None]
                else:
                    self.locs = [locs]
                self._data = [[array for array in inpt]]  # nested
                if locs is None and dims is None:
                    self.dims = [list(inpt[0].shape)[0]]
                else:
                    self.dims = dims
        else:
            raise ValueError('Unsupported data shape.')

    def tree_flatten(self):
        """Specifies a flattening recipe.

        Returns:
            a pair of an iterable with the children to be flattened recursively,
            and some opaque auxiliary data to pass back to the unflattening recipe.
            The auxiliary data is stored in the treedef for use during unflattening.
            The auxiliary data could be used, e.g., for dictionary keys.
        """
        return ((self.data, self.diag_unitary), (self.dims, self.locs))

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Specifies an unflattening recipe.

        Args:
            aux_data: the opaque data that was specified during flattening of the
            current treedef.
            children: the unflattened children

        Returns:
            a re-constructed object of the registered type, using the specified
            children and auxiliary data.
        """
        dims, locs = aux_data
        data, diag_unitary = children
        return cls(data,
                   dims,
                   locs,
                   _nested_inpt=True,
                   diag_unitary=diag_unitary)

    def get_data(self):
        return self._data

    def set_data(self, data):
        if not isinstance(data[0], list):
            raise TypeError('KronObj data must be in a nested list.')
        else:
            self._data = data

    data = property(get_data, set_data)

    def _build_sub_kronobj(self, idx):
        """Return a subset of `KronObj` at selected `idx`."""
        return KronObj(self.data[idx],
                       dims=self.dims,
                       locs=self.locs[idx],
                       diag_unitary=self.diag_unitary[idx])

    def _downcast_diagonal_unitary(self):
        """Convert the diagonal-unitary Hamiltonian format to dense matrix."""
        data = []
        for mat, eig_vec in zip(self.data, self.diag_unitary):
            if eig_vec is not None:  # downcasting
                if len(mat) == 1:
                    data.append([eig_vec * mat[0] @ jnp.conj(eig_vec).T])
                else:
                    raise TypeError('Unsupported diagonal-unitary format')
            else:
                data.append(mat)
        return KronObj(data, _nested_inpt=True, dims=self.dims, locs=self.locs)

    def _downcast_local_hamiltonian(self):
        """Cast the local Hamiltonian format to dense matrix."""
        # firstly downcast diag unitary
        if any(self.diag_status):
            self._downcast_diagonal_unitary()
        data = []
        for mats, local_list in zip(self.data, self.locs):
            if local_list is not None:  # downcasting
                # construct identity matrix list
                ilocs = [
                    idx for idx, _ in enumerate(self.dims)
                    if idx not in local_list
                ]
                idims = np.array(self.dims)[ilocs].prod()
                full_mat = tensor(*mats, jnp.eye(idims))
                tlocs = local_list + ilocs
                order = [tlocs.index(loc) for loc, _ in enumerate(self.dims)]
                # permute the full matrix in the correct order
                data.append([
                    permute(full_mat, [self.dims[loc] for loc in tlocs], order)
                ])
            else:
                data.append(mats)
        return KronObj(data, _nested_inpt=True, dims=self.dims)

    def __iter__(self):
        """iteration through nested list."""
        sublist = [
            self._build_sub_kronobj(ind) for ind, _ in enumerate(self.data)
        ]
        return iter(sublist)

    def __getitem__(self, key):

        if isinstance(key, int):
            return self._build_sub_kronobj(key)
        raise KeyError(f"Unrecognized key: {key}. Key must be an integer index")

    def __add__(self, other):
        """Addition with KronObj on left. The trivial addition is defined as
        concatenating the list.
        """
        if other == 0:  # identity of addition operator
            return self
        else:  # case for matching KronObj
            return KronObj(self.data + other.data,
                           self.dims,
                           _nested_inpt=True,
                           diag_unitary=self.diag_unitary + other.diag_unitary,
                           locs=self.locs + other.locs)

    def __radd__(self, other):
        """Addition with KronObj on right. Note the addition satisfies the
        commutative.
        """
        return self + other

    def __sub__(self, other):
        """Subtraction with KronObj on left."""
        return self + (-1.0 * other)

    def __rsub__(self, other):
        """Subtraction with KronObj on right."""
        return (-1.0 * self) + other

    def __mul__(self, other):
        """Multiplication with KronObj on left, the operator only acts on first
        term in the nested list. Note the scale multiplication satisfies the
        commutative.
        """
        if isinstance(other, (numbers.Number, jax.Array)):
            data = []
            for nest in self.data:
                # multiply scale number to first term
                array_0 = nest[0] * other
                data.append([array_0] + nest[1:])
            return KronObj(data,
                           self.dims,
                           _nested_inpt=True,
                           diag_unitary=self.diag_unitary,
                           locs=self.locs)
        else:
            raise NotImplementedError(f'Do not support instance {type(other)}.')

    def __rmul__(self, other):
        """Multiplication with KronObj on right"""
        return self * other

    def __matmul__(self, other):
        """Matrix multiplication with KronObj on left. The high efficient
        matrix-vector multiplication method working when `other` has matched shape.
        """
        new_self = self
        # downcast the diag unitary
        if any(self.diag_status):
            new_self = new_self._downcast_diagonal_unitary()
        if isinstance(other, KronObj):
            if all(self.loc_status) and all(other.loc_status) and not any(
                    self.diag_status) and not any(other.diag_status):
                # The product of two Kronecker products yields another Kronecker product
                res = []
                for kron_a in self:
                    for kron_b in other:
                        # concatenate loc info
                        dot_data = []
                        loc_info = sorted(set(kron_a.locs[0] + kron_b.locs[0]))
                        for loc in loc_info:
                            try:
                                id_a = kron_a.locs[0].index(loc)
                                mat_a = kron_a.data[0][id_a]
                            except ValueError:
                                mat_a = jnp.eye(kron_a.dims[loc])
                            try:
                                id_b = kron_b.locs[0].index(loc)
                                mat_b = kron_b.data[0][id_b]
                            except ValueError:
                                mat_b = jnp.eye(kron_b.dims[loc])
                            dot_data.append(mat_a @ mat_b)
                        # create new KronObj
                        res.append(KronObj(dot_data, self.dims, locs=loc_info))
                return sum(res)
            else:
                return self.full() @ other.full()
        elif isinstance(other, (np.ndarray, jnp.ndarray)):
            if other.shape[1] == 1:  # state vector
                if not all(new_self.loc_status):
                    return new_self.full() @ other
                res = []
                # implement Einstein summation
                for sub_kronobj in new_self:
                    # construct subscripts and operands
                    if len(sub_kronobj.data[0]) > 1:
                        unitary_op = tensor(*sub_kronobj.data[0])
                    else:
                        unitary_op = sub_kronobj.data[0][0]
                    locs = [loc + 1 for loc in sub_kronobj.locs[0]
                           ]  # index start from 1
                    aux_locs = [-loc for loc in locs]
                    unitary_dim = [
                        sub_kronobj.dims[idx] for idx in sub_kronobj.locs[0]
                    ]
                    unitary_op = unitary_op.reshape(unitary_dim * 2)
                    unitary_subscript = aux_locs + locs
                    psi_op = other.reshape(sub_kronobj.dims)
                    psi_subscript = list(np.arange(len(sub_kronobj.dims)) + 1)
                    res_subscript = psi_subscript.copy()
                    for loc in sub_kronobj.locs[0]:
                        res_subscript[loc] *= -1

                    new_psi = jnp.einsum(unitary_op, unitary_subscript, psi_op,
                                         psi_subscript,
                                         res_subscript).reshape(-1, 1)
                    res.append(new_psi)
                return sum(res)
            else:
                return new_self.full() @ other
        else:
            raise NotImplementedError(f'Do not support instance {type(other)}.')

    def __rmatmul__(self, other):
        """Matrix multiplication with KronObj on right"""
        if isinstance(other, (np.ndarray, jnp.ndarray)):
            return other @ self.full()
        else:
            raise NotImplementedError(f'Do not support instance {type(other)}.')

    def __str__(self):
        s = ""
        shape = self.shape
        s += ("Kronecker object: " + "dims = " + str(self.dims) + ", shape = " +
              str(shape) + ", diag_status = " + str(self.diag_status) +
              ", location_info = " + str(self.locs) + "\n")
        s += "KronObj data =\n"
        s += str(self.data)

        return s

    def __repr__(self):
        # give complete information on KronObj without print statement in
        # command-line we cant realistically serialize a KronObj into a string,
        # so we simply return the informal __str__ representation instead.)
        return self.__str__()

    @property
    def shape(self):
        """The shape of unitary"""
        res = [np.prod(self.dims), np.prod(self.dims)]
        return res

    @property
    def diag_status(self):
        """The diagonal-unitary format status of subsystems."""
        return [eig_vec is not None for eig_vec in self.diag_unitary]

    @property
    def loc_status(self):
        """The local Hamiltonian format status of subsystems."""
        return [loc_info is not None for loc_info in self.locs]

    def full(self):
        """Dense matrix from the result of Kronecker product."""
        new_self = self
        # downcast the diag unitary
        if any(self.diag_status):
            new_self = new_self._downcast_diagonal_unitary()
        # downcast the location representation
        if any(self.loc_status):
            new_self = new_self._downcast_local_hamiltonian()
        mat_list = [mats[0] for mats in new_self.data]
        return sum(mat_list)

    def dense(self):
        """Return new KronObj contain dense matrix"""
        data = [self.full()]
        return KronObj(data,
                       dims=self.dims,
                       locs=[i for i, _ in enumerate(self.dims)])

    def dag(self):
        """Conjugate transpose over tensor product"""
        data = []
        for nest_list in self.data:
            new_nest_list = []
            for mat in nest_list:
                new_nest_list.append(mat.conj().T)
            data.append(new_nest_list)
        return KronObj(data,
                       _nested_inpt=True,
                       dims=self.dims,
                       locs=self.locs,
                       diag_unitary=self.diag_unitary)

    def transpose(self):
        """Transpose over tensor product"""
        data = []
        diag_unitary = []
        for nest_list in self.data:
            new_nest_list = []
            for mat in nest_list:
                new_nest_list.append(mat.T)
            data.append(new_nest_list)
        for eig_vec in self.diag_unitary:
            if eig_vec is not None:
                diag_unitary.append(eig_vec.conj())
            else:
                diag_unitary.append(None)
        return KronObj(data,
                       _nested_inpt=True,
                       dims=self.dims,
                       locs=self.locs,
                       diag_unitary=diag_unitary)

    def conjugate(self):
        """Conjugate over tensor product"""
        data = []
        diag_unitary = []
        for nest_list in self.data:
            new_nest_list = []
            for mat in nest_list:
                new_nest_list.append(mat.conj())
            data.append(new_nest_list)
        for eig_vec in self.diag_unitary:
            if eig_vec is not None:
                diag_unitary.append(eig_vec.conj())
            else:
                diag_unitary.append(None)
        return KronObj(data,
                       _nested_inpt=True,
                       dims=self.dims,
                       locs=self.locs,
                       diag_unitary=diag_unitary)

    def diagonalize_operator(self):
        """Diagonalize Hermitian to diagonal-unitary representation."""
        data = []
        diag_unitary = []
        for kron_region in self.data:
            mat = tensor(*kron_region)
            # we cannot diagonalize a non-Hermite matrix
            # Temporarily comment it out to enhance jit support, you must ensure
            # the input is Hermitian to use diag evolution method.

            # if not np.allclose(np.conj(mat).T, mat):
            #     continue
            eig, eigv = jaxLA.eigh(mat)
            data.append([eig])
            diag_unitary.append(
                jax.lax.stop_gradient(eigv))  # Fix complex casting bug
        return KronObj(data,
                       _nested_inpt=True,
                       dims=self.dims,
                       locs=self.locs,
                       diag_unitary=diag_unitary)

    def compute_contraction_path(self, op_list=None, trotter_order=None):
        """Compute tensor network contraction path.
        Please provide either `op_list` or `trotter_order`. If `op_list` is None,
        it will be generated by `trotter_order`.

        Args:
            op_list: the list of local operator
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
        """
        assert op_list is not None or trotter_order is not None
        if op_list is None:
            op_list = self._trotter_decomposition(trotter_order)
        # initialize subscripts by state
        max_indices_per_layer = len(self.dims)
        num_subscript = np.zeros(len(self.dims), dtype=int)
        psi_subscript = ''.join([
            oe.get_symbol(i + max_indices_per_layer * num)
            for i, num in enumerate(num_subscript)
        ])
        oe_subscripts = []
        oe_shapes = []
        # construct tensor network
        oe_subscripts.append(psi_subscript)
        oe_shapes.append(self.dims)
        for sub_kronobj in op_list[::-1]:
            base_subscript = []
            new_subscript = []
            op_shape = []
            for loc in sub_kronobj.locs[0]:
                base_subscript.append(
                    oe.get_symbol(loc +
                                  max_indices_per_layer * num_subscript[loc]))
                num_subscript[loc] += 1
                new_subscript.append(
                    oe.get_symbol(loc +
                                  max_indices_per_layer * num_subscript[loc]))
                op_shape.append(self.dims[loc])
            oe_subscripts.append(''.join(new_subscript + base_subscript))
            oe_shapes.append(op_shape * 2)
        psi_output = ''.join([
            oe.get_symbol(i + max_indices_per_layer * num)
            for i, num in enumerate(num_subscript)
        ])
        eq = ','.join(oe_subscripts) + '->' + psi_output
        expr = oe.contract_expression(eq, *oe_shapes, optimize='auto-hq')
        return expr

    def _trotter_decomposition(self, trotter_order):
        """Generate Suzuki-Trotter decomposition sequence.

        Args:
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
        """

        def _loop(trotter_coeff, reverse=False):
            """Generate the matrix multiplication sequence.

            Args:
                trotter_coeff (float): the trotter coefficient
                reverse (bool): whether reverse the sequence or not.
            """
            op_list = []
            if reverse:
                scan_list = zip(reversed(self.data),
                                reversed(self.diag_unitary),
                                reversed(self.locs))
            else:
                scan_list = zip(self.data, self.diag_unitary, self.locs)
            for mat_list, eig_vec, local_info in scan_list:
                if eig_vec is not None:  # cast expm down to exp
                    lam = jnp.exp(trotter_coeff * mat_list[0])
                    op_list.append(
                        KronObj([eig_vec * lam @ jnp.conj(eig_vec).T],
                                self.dims, local_info))
                else:
                    # only calculate matrix exponentiation with the terms in `local_info`
                    obj = tensor(*mat_list) * trotter_coeff
                    op_list.append(
                        KronObj([jaxLA.expm(obj)], self.dims, local_info))
            return op_list

        def _second_order_expm(p=1.0):
            """Generate the matrix multiplication sequence for second order
            trotter decomposition.
            """
            # normal loop
            op_list = _loop(p / 2., reverse=False)
            # reversed loop
            op_list.extend(_loop(p / 2., reverse=True))
            return op_list

        if trotter_order == 1:
            op_list = _loop(1.)
        elif trotter_order == 2:
            op_list = _second_order_expm()
        elif trotter_order == 4:
            # the simplest real decomposition of 4th order STD
            p = 1 / (2 - 2**(1 / 3))
            op_list = _second_order_expm(p)
            op_list.extend(_second_order_expm(1 - 2 * p))
            op_list.extend(_second_order_expm(p))
        elif trotter_order == 4j:
            # the complex 4th STD
            p = (3 - 1j * sqrt(3)) / 6
            p_conj = (3 + 1j * sqrt(3)) / 6
            op_list = _second_order_expm(p)
            op_list.extend(_second_order_expm(p_conj))
        else:
            raise ValueError(f'Unsupported Trotter order {trotter_order}.')

        return op_list

    def expm(self, right_vec=None, trotter_order=None, tn_expr=None):
        """Calculate matrix exponentiation of `KronObj`. Combine matrix exponentiation
        with matrix-vector product if `right_vec` is not `None`.

        Args:
            right_vec (bool): matrix-vector product with the `right_vec` on the right
            trotter_order (complex int): the order of suzuki-trotter decomposition.
                The following arguments are supported,
                a) `None`, calculating matrix exponentiation without trotter decomposition
                b) `1`, first order trotter decomposition
                c) `2`, second order trotter decomposition
                d) `4`, 4th order real decomposition
                e) `4j`, 4th order complex decomposition
            tn_expr: Tensor network contract expression.
                if `tn_expr` is None, generate contract expression when `expm()`
                be called each time.

        """
        if right_vec is None or trotter_order not in [1, 2, 4, 4j]:
            # not implement trotter approximate
            res = jaxLA.expm(self.full())
            if right_vec is not None:
                return res @ right_vec
            else:
                return res
        else:
            op_list = self._trotter_decomposition(trotter_order)
            if tn_expr is None:
                tn_expr = self.compute_contraction_path(op_list)
            oe_arrays = []
            oe_arrays.append(right_vec.reshape(self.dims))
            for sub_kronobj in op_list[::-1]:
                op_dim = [sub_kronobj.dims[idx] for idx in sub_kronobj.locs[0]]
                op_array = sub_kronobj.data[0][0].reshape(op_dim * 2)
                oe_arrays.append(op_array)
            oe_res = tn_expr(*oe_arrays, backend='jax')

            return oe_res.reshape(-1, 1)



import numpy as np
from supergrad.utils import (gates, compute_fidelity_with_1q_rotation_axis)


def test_fidelity_vz():
    ut = gates.cz_gate(2)
    angle = np.array([0.78129602, 1.42407966, -2.30043467, 1.3672363])
    # angle = np.array([0.7, 1.4, -2.3, 1.54])
    sim_u = np.diag(np.exp(1j * angle))
    rng = np.random.default_rng(2983461)
    sim_u += rng.uniform(-1e-4, 1e-4, 4 * 4).reshape((4, 4))
    for method in ["gd", "lbfgs", "simple_vz", "gd_init2", "lbfgs_init2"]:
        fid, _ = compute_fidelity_with_1q_rotation_axis(u_target=ut,
                                                        u_computed=sim_u,
                                                        opt_method=method,
                                                        compensation_option="only_vz")
        assert fid > 0 and fid < 1
